# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from importlib import import_module

from climetlab.core import Base
from climetlab.decorators import locked

LOG = logging.getLogger(__name__)


class Wrapper(Base):
    pass


_HELPERS = {}


# TODO: Add plugins
@locked
def _wrappers():
    if not _HELPERS:
        here = os.path.dirname(__file__)
        for path in os.listdir(here):
            if path.endswith(".py") and path[0] not in ("_", "."):
                name, _ = os.path.splitext(path)
                try:
                    _HELPERS[name] = import_module(f".{name}", package=__name__).wrapper
                except Exception as e:
                    LOG.warning(f"Error loading wrapper '{name}': {e}")
    return _HELPERS


def get_wrapper(data, *args, **kwargs):
    """
    Returns an object that wraps classes from other packages
    to support
    """

    if isinstance(data, Base):
        return data

    for name, h in _wrappers().items():
        wrapper = h(data, *args, **kwargs)
        if wrapper is not None:
            return wrapper.mutate()

    fullname = ".".join([data.__class__.__module__, data.__class__.__qualname__])

    raise ValueError(f"Cannot find a wrapper for class {fullname}")



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab import load_source
from climetlab.utils.dates import to_datetime
from climetlab.utils.domains import domain_to_area

from . import Dataset


class Era5SingleLevels(Dataset):
    def __init__(self, variable, period, domain=None, time=None, grid=None):
        self.variable = variable

        request = dict(
            variable=self.variable,
            product_type="reanalysis",
        )

        if domain is not None:
            request["area"] = domain_to_area(domain)

        if time is not None:
            request["time"] = time
        else:
            request["time"] = list(range(0, 24))

        if grid is not None:
            if isinstance(grid, (int, float)):
                request["grid"] = [grid, grid]
            else:
                request["grid"] = grid

        if isinstance(period, int):
            period = (period, period)

        if isinstance(period, (tuple, list)) and len(period) == 1:
            period = (period[0], period[0])

        sources = []
        for year in range(period[0], period[1] + 1):
            request["year"] = year
            sources.append(
                load_source("cds", "reanalysis-era5-single-levels", **request)
            )

        self.source = load_source("multi", sources)

    def __getitem__(self, n):
        # TODO: move to superclass
        if isinstance(n, int):
            return super().__getitem__(n)

        n = to_datetime(n)
        for field in self:
            if field.datetime() == n:
                return field

        raise KeyError(n)


dataset = Era5SingleLevels



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from climetlab.readers import Reader
from climetlab.readers.grib.index import GribMultiFieldSet
from climetlab.readers.grib.index.file import FieldSetInOneFile

LOG = logging.getLogger(__name__)


class GRIBReader(FieldSetInOneFile, Reader):
    appendable = True  # GRIB messages can be added to the same file

    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        FieldSetInOneFile.__init__(self, path)

    def __repr__(self):
        return "GRIBReader(%s)" % (self.path,)

    @classmethod
    def merge(cls, readers):
        assert all(isinstance(s, GRIBReader) for s in readers), readers
        assert len(readers) > 1

        return GribMultiFieldSet(readers)

    def mutate_source(self):
        # A GRIBReader is a source itself
        return self



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import calendar
import datetime
import logging
import threading

import climetlab as cml
from climetlab.readers.grib.index import GribFieldSet
from climetlab.utils.serialise import register_serialisation

LOG = logging.getLogger(__name__)


class NoLock:
    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass


class VirtualField:
    def __init__(self, h, owner):
        self.owner = owner
        date = datetime.datetime(1959, 1, 1) + datetime.timedelta(hours=h)
        self.date = date.year * 10000 + date.month * 100 + date.day
        self.time = date.hour
        self.index = (self.date % 100 - 1) * 24 + self.time

    def metadata(self, n):
        if n == "dataDate":
            return self.date

        if n == "dataTime":
            return self.time

        return self.owner.reference[n]

    @property
    def values(self):
        source = self.owner.full_month(self.date)
        return source[self.index].values

    def to_numpy(self):
        source = self.owner.full_month(self.date)
        return source[self.index].to_numpy()

    @property
    def shape(self):
        source = self.owner.full_month(self.date)
        return source[self.index].shape


class DictOveray(dict):
    def __init__(self, field):
        self.field = field

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            self[key] = self.field[key]
            return self[key]


class Virtual(GribFieldSet):
    SIZE = int(365.25 * 24 * (2022 - 1959))
    # SIZE = 100

    def __init__(self, **kwargs):
        super().__init__()
        self.request = dict(
            dataset="reanalysis-era5-single-levels",
            product_type="reanalysis",
            param="2t",
            grid="10/10",
        )
        self.request.update(kwargs)

        self.reference = DictOveray(
            cml.load_source(
                "cds",
                date=19590101,
                time=0,
                **self.request,
            )[0]
        )

        self.fields = {}
        self.lock = threading.Lock()
        self.locks = {}

    def __len__(self):
        return self.SIZE

    def __getitem__(self, i):
        if i >= self.SIZE:
            raise IndexError()

        return VirtualField(i, self)

    def xarray_open_dataset_kwargs(self):
        return dict(
            cache=False,  # Set to false to prevent loading the whole dataset
            chunks={
                "time": 24 * 31,
                # "step": 1,
                # "number": 1,
                # "surface": 1,
                "latitude": 721,
                "longitude": 1440,
            },
            lock=NoLock(),
        )

    def full_month(self, date):
        yyyymm = date // 100

        with self.lock:
            if yyyymm in self.fields:
                return self.fields[yyyymm]
            if yyyymm not in self.locks:
                self.locks[yyyymm] = threading.Lock()

        with self.locks[yyyymm]:
            # Some other threads may have created that in the meantime

            if yyyymm in self.fields:
                return self.fields[yyyymm]

            yyyy = yyyymm // 100
            mm = yyyymm % 100
            last = calendar.monthrange(yyyy, mm)[1]
            source = cml.load_source(
                "cds",
                year=yyyy,
                month=mm,
                day=list(range(1, last + 1)),
                time=list(range(0, 24)),
                **self.request,
            )
            self.fields[yyyymm] = source
            return source


register_serialisation(Virtual, lambda x: None, lambda x: Virtual())


source = Virtual



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.indexing.fieldset import FieldSet


class EmptySource(FieldSet):
    def ignore(self):
        # Used by multi-source
        return True

    def __len__(self):
        return 0

    def __repr__(self):
        return "EmptySource"


source = EmptySource



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import glob
import logging
import os
import time

import pyfdb
import yaml

from climetlab.readers.grib.index import FieldSetInFiles
from climetlab.utils.parts import Part

LOG = logging.getLogger(__name__)


class FDB(FieldSetInFiles):
    def __init__(self, root=None, schema=None, request={}):
        super().__init__(db=None)

        if schema is None and root is not None:
            for n in glob.iglob(f"{root}/*/schema"):
                schema = n
                break

        config = {
            "spaces": [{"roots": [{"path": root}]}],
            "schema": schema,
        }
        os.environ["FDB5_CONFIG"] = yaml.dump(config)

        now = time.time()

        fdb = pyfdb.FDB()
        self.parts = list(fdb.list(request))
        print("pyfdb.list", time.time() - now)

    def number_of_parts(self):
        return len(self.parts)

    def part(self, i):
        f = self.parts[i]
        return Part(f["path"], f["offset"], f["length"])


source = FDB



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from climetlab.wrappers import get_wrapper

from . import Reader

LOG = logging.getLogger(__name__)


class FWFReader(Reader):
    def __init__(self, source, path, header=None):
        super().__init__(source, path)
        self.header = header

    def to_pandas(self, **kwargs):
        import pandas

        pandas_read_fwf_kwargs = kwargs.get("pandas_read_fwf_kwargs", {})

        LOG.debug(
            "pandas.read_fwf(%s,header=%s,%s)",
            self.path,
            self.header,
            pandas_read_fwf_kwargs,
        )
        return pandas.read_fwf(
            self.path,
            header=self.header,
            **pandas_read_fwf_kwargs,
        )

    def plot_map(self, backend):
        get_wrapper(self.to_pandas()).plot_map(backend)


def reader(source, path, magic=None, deeper_check=False):
    if magic is None:  # Bypass check and force
        return FWFReader(source, path)

    # For now, only created if magic is None


aliases = ["fix_width_format"]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from datetime import timedelta
from functools import cached_property

import numpy as np

from climetlab.indexing.fieldset import Field
from climetlab.utils.bbox import BoundingBox

from .coords import LevelSlice, TimeSlice
from .dataset import DataSet


class NetCDFField(Field):
    def __init__(self, owner, ds, variable, slices, non_dim_coords):
        data_array = ds[variable]

        self.owner = owner
        self.variable = variable
        self.slices = slices
        self.non_dim_coords = non_dim_coords
        self.shape = (data_array.shape[-2], data_array.shape[-1])
        self.ds = ds

        self.name = self.variable

        self.title = getattr(
            data_array,
            "long_name",
            getattr(data_array, "standard_name", self.variable),
        )

        self.time = non_dim_coords.get("valid_time", non_dim_coords.get("time"))
        self.level = None
        self.levtype = "sfc"

        for s in self.slices:
            if isinstance(s, TimeSlice):
                self.time = s.value

            if s.is_info:
                self.title += " (" + s.name + "=" + str(s.value) + ")"

            if isinstance(s, LevelSlice):
                self.level = s.value
                self.levtype = {"pressure": "pl"}.get(s.name, s.name)

        if "forecast_reference_time" in ds.data_vars:
            forecast_reference_time = ds["forecast_reference_time"].data
            assert forecast_reference_time.ndim == 0, forecast_reference_time
            forecast_reference_time = forecast_reference_time.astype("datetime64[s]")
            forecast_reference_time = forecast_reference_time.astype(object)
            step = (self.time - forecast_reference_time).total_seconds()
            assert step % 3600 == 0, step
            self.step = int(step // 3600)

    def plot_map(self, backend):
        dimensions = dict((s.name, s.index) for s in self.slices)

        backend.bounding_box(
            north=self.north, south=self.south, west=self.west, east=self.east
        )

        backend.plot_netcdf(self.owner.path, self.variable, dimensions)

    def __repr__(self):
        return "NetCDFField[%r,%r]" % (self.variable, self.slices)

    def to_bounding_box(self):
        return BoundingBox(
            north=self.north,
            south=self.south,
            east=self.east,
            west=self.west,
        )

    @cached_property
    def projection(self):
        def tidy(x):
            if isinstance(x, np.ndarray):
                return x.tolist()
            if isinstance(x, (tuple, list)):
                return [tidy(y) for y in x]
            if isinstance(x, dict):
                return {k: tidy(v) for k, v in x.items()}
            return x

        return tidy(
            self.owner.xr_dataset[
                self.owner.xr_dataset[self.variable].grid_mapping
            ].attrs
        )

    # Compatibility to GRIb fields below

    def grid_points(self):
        return DataSet(self.owner.xr_dataset).grid_points(self.variable)

    def grid_points_xy(self):
        return DataSet(self.owner.xr_dataset).grid_points_xy(self.variable)

    def to_numpy(self, reshape=True, dtype=None):
        dimensions = dict((s.name, s.index) for s in self.slices)
        values = self.owner.xr_dataset[self.variable].isel(dimensions).values
        if not reshape:
            values = values.flatten()
        if dtype is not None:
            values = values.astype(dtype)
        return values

    # ----------------------------------------------------------------
    def metadata(self, name):
        return getattr(self, f"_metadata_{name}")()

    def _metadata_valid_datetime(self):
        return self.time.isoformat()

    def _metadata_param(self):
        return self.variable

    def _metadata_step(self):
        return self.step

    def _metadata_levelist(self):
        return self.level

    def _metadata_levtype(self):
        return self.levtype

    def _metadata_number(self):
        # TODO: code me
        return None

    def _metadata_date(self):
        date = self.time - timedelta(hours=self.step)
        return int(date.strftime("%Y%m%d"))

    def _metadata_time(self):
        date = self.time - timedelta(hours=self.step)
        return int(date.strftime("%H%M"))

    # ----------------------------------------------------------------
    def as_mars(self):
        return dict(
            param=self._metadata_param(),
            step=self._metadata_step(),
            levelist=self._metadata_levelist(),
            levtype=self._metadata_levtype(),
            number=self._metadata_number(),
            date=self._metadata_date(),
            time=self._metadata_time(),
        )

    @property
    def mars_area(self):
        # TODO: code me
        return [self.north, self.west, self.south, self.east]

    @property
    def mars_grid(self):
        # TODO: code me
        return None

    # ----------------------------------------------------------------
    @property
    def resolution(self):
        # TODO: code me
        return None

    def valid_datetime(self):
        return self.time.isoformat()

    @property
    def north(self):
        return self.ds.bbox(self.variable)[0]

    @property
    def west(self):
        return self.ds.bbox(self.variable)[1]

    @property
    def south(self):
        return self.ds.bbox(self.variable)[2]

    @property
    def east(self):
        return self.ds.bbox(self.variable)[3]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import os
from collections import defaultdict

import yaml

from climetlab.decorators import locked

MAGICS_KEYS = None
MAGICS_DEF = None
MAGICS_PARAMS = None
_inited = False


@locked
def init():
    global _inited, MAGICS_KEYS, MAGICS_DEF, MAGICS_PARAMS

    if not _inited:
        MAGICS_KEYS = defaultdict(set)
        MAGICS_PARAMS = defaultdict(dict)
        with open(os.path.join(os.path.dirname(__file__), "magics.yaml")) as f:
            MAGICS_DEF = yaml.load(f, Loader=yaml.SafeLoader)
            for action, params in MAGICS_DEF.items():
                for param in params:
                    name = param["name"]
                    MAGICS_KEYS[name].add(action)
                    MAGICS_PARAMS[action][name] = param

    _inited = True


def magics_keys_to_actions():
    init()
    return MAGICS_KEYS


def magics_keys_definitions():
    init()
    return MAGICS_DEF


def magics_keys_parameters(name):
    init()
    return MAGICS_PARAMS[name]



#!/usr/bin/env python3
# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import warnings
from itertools import zip_longest
from numbers import Number

import numpy as np


def default_merger(*funcs):
    if not funcs:
        return None

    def map_fn(i):
        i = int(i)
        arrays = [m(i) for m in funcs]
        array = np.stack(arrays)
        return array

    return map_fn


def as_numpy_func(ds, options=None):
    if ds is None or callable(ds):
        return ds

    def _options(new):
        o = {k: v for k, v in ds.get_options().items()}
        if new:
            o.update(new)
        return o

    options = _options(options)

    to_numpy_kwargs = options.get("to_numpy_kwargs", {})

    def take_i(i):
        return ds[i].to_numpy(**to_numpy_kwargs)

    if "offset" in options and options["offset"]:
        offset = options["offset"]

        def take_i(i):  # noqa: F811
            return ds[i + offset].to_numpy(**to_numpy_kwargs)

    func = take_i

    if "constant" in options and options["constant"]:

        def first(func):
            def wrap(i):
                return func(0)

            return wrap

        func = first(func)

    if "normalize" in options:
        a, b = normalize_a_b(options["normalize"], ds)

        def normalize(func):
            def wrap(i):
                return a * func(i) + b

            return wrap

        func = normalize(func)

    return func


def normalize_a_b(option, dataset):
    if isinstance(option, (tuple, list)) and all(
        [isinstance(x, Number) for x in option]
    ):
        a, b = option
        return a, b

    if option == "mean-std":
        stats = dataset.statistics()
        average, stdev = stats["average"], stats["stdev"]
        if stdev < (average * 1e-6):
            warnings.warn(
                f"Normalizing: the field seems to have only one value {stats}"
            )
        return 1 / stdev, -average / stdev

    if option == "min-max":
        stats = dataset.statistics()
        mini, maxi = stats["minimum"], stats["maximum"]
        x = maxi - mini
        if x < 1e-9:
            warnings.warn(
                f"Normalizing: the field seems to have only one value {stats}."
            )
        return 1 / x, -mini / x

    raise ValueError(option)


def to_funcs(features, targets, options, targets_options, merger, targets_merger):
    if targets is None:
        targets = []

    if options is None:
        options = [{} for i in features]

    if targets_options is None:
        targets_options = [{} for i in targets]

    assert isinstance(features, (list, tuple)), features
    assert len(features) == len(options), (len(features), len(options))
    funcs = [
        as_numpy_func(_, opt) for _, opt in zip_longest(features, options, fillvalue={})
    ]
    funcs_targets = [
        as_numpy_func(_, opt)
        for _, opt in zip_longest(targets, targets_options, fillvalue={})
    ]

    func = merger(*funcs)
    func_targets = targets_merger(*funcs_targets)

    return func, func_targets



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from climetlab.sources.file import FileSource
from climetlab.utils import string_to_args

LOG = logging.getLogger(__name__)

FORWARDS = ("to_xarray", "to_pandas", "to_tfdataset")


def _nearest_common_class(objects):
    # mro() is "method resolution order"
    mros = [type(o).mro() for o in objects]

    first = mros[0]
    rest = mros[1:]
    for c in first:
        if all(c in m for m in rest):
            return c

    assert False


def _flatten(sources):
    from climetlab.sources.multi import MultiSource

    for s in sources:
        if isinstance(s, MultiSource) and s.merger is None:
            yield from _flatten(s.sources)
        else:
            yield s


def merge_by_class(sources):
    common = _nearest_common_class(sources)
    return common.merge(sources)


class Merger:
    def __init__(self, sources):
        assert sources

        self.sources = list(_flatten(sources))
        assert self.sources, sources

        self.paths = None
        self.reader_class = None
        self.common = _nearest_common_class(sources)
        LOG.debug("nearest_common_class %s", self.common)

        if issubclass(self.common, FileSource):
            readers = [s._reader for s in self.sources]
            self.reader_class = _nearest_common_class(readers)
            LOG.debug("nearest_common_class %s", self.reader_class)
            self.paths = [s.path for s in self.sources]

    @property
    def paths_or_sources(self):
        if self.paths is not None:
            return self.paths
        return self.sources


class DefaultMerger(Merger):
    def to_pandas(self, **kwargs):
        from .pandas import merge

        return merge(
            sources=self.sources,
            paths=self.paths,
            reader_class=self.reader_class,
            **kwargs,
        )

    def to_tfdataset(self, **kwargs):
        from .tfdataset import merge

        return merge(
            sources=self.sources,
            paths=self.paths,
            reader_class=self.reader_class,
            **kwargs,
        )

    def to_xarray(self, **kwargs):
        from .xarray import merge

        return merge(
            sources=self.sources,
            paths=self.paths,
            reader_class=self.reader_class,
            **kwargs,
        )


class ObjMerger(Merger):
    def __init__(self, obj, sources, *args, **kwargs):
        super().__init__(sources)
        self.obj = obj

    def to_xarray(self, *args, **kwargs):
        return self.obj.to_xarray(self.paths_or_sources, **kwargs)

    def to_pandas(self, *args, **kwargs):
        return self.obj.to_pandas(self.paths_or_sources, **kwargs)

    def to_tfdataset(self, *args, **kwargs):
        return self.obj.to_tfdataset(self.paths_or_sources, **kwargs)


class CallableMerger(Merger):
    def __init__(self, func, sources, *args, **kwargs):
        super().__init__(sources)
        self.func = func

    def _call_func(self, *args, **kwargs):
        return self.func(self.paths_or_sources, **kwargs)

    to_xarray = _call_func
    to_pandas = _call_func
    to_tfdataset = _call_func


class XarrayGenericMerger(Merger):
    def __init__(self, sources, **options):
        super().__init__(sources)
        self.options = options

    def to_xarray(self, *args, **kwargs):
        assert self.paths is not None, self.paths
        import xarray as xr

        options = {}
        options.update(self.default_options)
        options.update(self.options)
        options.update(kwargs)
        LOG.debug(f"xr.open_mfdataset with options = {options}")
        return xr.open_mfdataset(
            self.paths,
            **options,
        )


class XarrayConcatMerger(XarrayGenericMerger):
    def __init__(self, sources, **options):
        if "dim" in options:
            dim = options.pop("dim")
            options["concat_dim"] = dim
        super().__init__(sources, **options)

    default_options = {"combine": "nested"}


class XarrayMerger(XarrayGenericMerger):
    default_options = {}


MERGERS = {
    "concat": XarrayConcatMerger,
    "merge": DefaultMerger,
}


def add_default_values_and_kwargs(args):
    kwargs = dict()
    for a in args:
        k, v = a.split("=")
        kwargs[k] = v
    return kwargs


def make_merger(merger, sources):
    for fwd in FORWARDS:
        if hasattr(merger, fwd) and callable(getattr(merger, fwd)):
            LOG.debug("Merger %s has method in %s()", merger, fwd)
            return ObjMerger(merger, sources)

    if callable(merger):
        LOG.debug("Merger %s is callable", merger)
        return CallableMerger(merger, sources)

    if isinstance(merger, str):
        name, args, kwargs = string_to_args(merger)
        return MERGERS[name](sources, *args, **kwargs)

    if isinstance(merger, tuple):
        if len(merger) == 2 and isinstance(merger[1], dict):
            return MERGERS[merger[0]](sources, **merger[1])
        return MERGERS[merger[0]](sources, *merger[1:])

    if merger is None:
        LOG.debug("Using DefaultMerger")
        return DefaultMerger(sources)

    raise ValueError(f"Unsupported merger {merger} ({type(merger)})")



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import datetime
import json
import logging
import os

from climetlab.utils.parts import Part

from . import Database

LOG = logging.getLogger(__name__)


class JsonDatabase(Database):
    VERSION = 1

    def __init__(self, *arg):
        pass

    def json_dumps_entry(self, entry):
        return json.dumps(entry, default=json_serialiser)


class JsonStdoutDatabase(JsonDatabase):
    def load_iterator(self, iterator):
        count = 0
        for entry in iterator:
            print(self.json_dumps_entry(entry))
            count += 1
        return count


def json_serialiser(o):
    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()
    if hasattr(o, "as_dict"):
        return o.as_dict()


class JsonFileDatabase(JsonDatabase):
    EXTENSION = ".json"

    def __init__(self, db_path):
        self.entries = []
        self.db_path = db_path
        if os.path.exists(db_path):
            with open(db_path) as f:
                for entry in f:
                    self.entries.append(json.loads(entry))

    def load_iterator(self, iterator):
        self.entries = []

        with open(self.db_path, "w") as f:
            for entry in iterator:
                self.entries.append(entry)
                print(self.json_dumps_entry(entry), file=f)

        return len(self.entries)

    def lookup_parts(self, order=None, **kwargs):
        parts = []
        for e in self.entries:
            parts.append(Part(e["_path"], e["_offset"], e["_length"]))
        return Part.resolve(parts, os.path.dirname(self.db_path))

    def lookup_dicts(self, *args, **kwargs):
        return self.entries



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse
import functools
import shlex
import textwrap
from itertools import cycle

from termcolor import colored


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ValueError(f"{self.prog}: {message}\n\n{self.format_help()}.")


def parse_args(epilog="", **kwargs):
    def wrapper(func):
        doc = "" if func.__doc__ is None else func.__doc__

        p = ArgumentParser(
            func.__name__.replace("do_", ""),
            description=textwrap.dedent(doc),
            epilog=textwrap.dedent(epilog),
            add_help=False,
        )
        # custom help to avoid exiting from climetlab cli.
        p.add_argument(
            "-h", "--help", action="store_true", help="show this help message and exit"
        )

        for k, v in kwargs.items():
            k = k.replace("_", "-")

            if isinstance(v, dict):  # short form is handled directly
                p.add_argument(f"--{k}", **v)
                continue

            assert isinstance(v, (list, tuple)), v
            lst, dic = v[:-1], v[-1]
            assert isinstance(dic, dict), dic
            lst = list(lst)
            if len(lst) == 0:
                lst = [k]
            if lst[0] is None:
                lst[0] = k
            if not lst[0] in [k, f"--{k}"]:
                lst = [f"--{k}"] + lst

            p.add_argument(*lst, **dic)

        func._argparser = p
        func._kwargs_specifications = kwargs

        @functools.wraps(func)
        def wrapped(self, args):
            args = shlex.split(args)  # shlex honors the quotes
            args = p.parse_args(args)
            if args.help:
                p.print_help()
                return
            return func(self, args)

        return wrapped

    return wrapper


def experimental(func):
    func._climetlab_experimental = True
    return func


def print_table(rows, colours=["blue"]):
    rows = list(rows)
    colours = list(colours)
    width = max(len(x[0]) for x in rows)
    for row, colour in zip(rows, cycle(colours)):
        print("{0:<{width}} {1}".format(row[0], colored(row[1], colour), width=width))



import logging
import threading

LOG = logging.getLogger(__name__)


class Future:
    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._condition = threading.Condition()
        self._ready = False
        self._result = None

    def execute(self):
        try:
            self._result = self.func(*self.args, **self.kwargs)
        except Exception as e:
            LOG.error(e)
            self._result = e
        with self._condition:
            self._ready = True
            self._condition.notify_all()

    def result(self):
        with self._condition:
            while not self._ready:
                self._condition.wait()
        if isinstance(self._result, Exception):
            raise self._result
        return self._result


class SoftThread(threading.Thread):
    def __init__(self, queue, condition):
        super().__init__(daemon=True)
        self._queue = queue
        self._condition = condition

    def run(self):
        while True:
            with self._condition:
                while len(self._queue) == 0:
                    self._condition.wait()
                s = self._queue.pop(0)
                self._condition.notify_all()
            if s is None:
                break
            s.execute()


class SoftThreadPool:
    def __init__(self, nthreads=4):
        self._nthreads = nthreads
        self._queue = []
        self._condition = threading.Condition()
        self._active = False
        self._lock = threading.RLock()
        self._error = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(exc_type)

    def start(self):
        with self._lock:
            self._active = True
            for _ in range(self._nthreads):
                SoftThread(self._queue, self._condition).start()

    def shutdown(self, error=None):
        with self._lock:
            if not self._active:
                return

        for _ in range(self._nthreads):
            with self._condition:
                self._queue.append(None)
                self._condition.notify_all()

        with self._lock:
            self._active = False
            self._error = error

    def submit(self, func, *args, **kwargs):  # submit
        with self._lock:
            if not self._active:
                self.start()

        with self._condition:
            s = Future(func, args, kwargs)
            self._queue.append(s)
            self._condition.notify_all()
            return s

    def __call__(self):
        with self._lock:
            if self._error:
                raise self._error



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from climetlab.decorators import normalize
from climetlab.sources import Source

LOG = logging.getLogger(__name__)


class IndexedSource(Source):
    _reader_ = None

    @normalize("date", "date-list", format="%Y%m%d")
    def __init__(self, index, order_by=None, filter=None, merger=None, **kwargs):
        LOG.debug(f"New IndexedSource order={order_by} kwargs={kwargs}")

        def _build_order_by_from_selection(selection):
            if not selection:
                return None
            order_by = {}
            for k, v in selection.items():
                if v is None:
                    order_by[k] = v
                elif isinstance(v, (list, tuple)):
                    order_by[k] = v
                else:
                    # There is only one element for this key in the request.
                    # No need to sort along this dimension
                    pass
                    # TODO: remove this comment:
                    # We could still keep this key to ensure consistency: when coming from a request,
                    # the values of order_by would be either None or a list with the following.
                    #   either:
                    #      order_by[k] = [v]
                    #   or:
                    #      order_by[k] = None
            return order_by

        if order_by is None:
            order_by = _build_order_by_from_selection(kwargs)

        self.filter = filter
        self.merger = merger
        self.index = index

        if kwargs:
            self.index = self.index.sel(kwargs)

        if order_by:
            self.index = self.index.order_by(order_by)

        super().__init__()

    def mutate(self):
        return self.index



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from climetlab.readers.grib.index import FieldsetInFilesWithJsonIndex
from climetlab.sources.indexed import IndexedSource


class IndexedUrlWithJsonIndex(IndexedSource):
    # This class is mostly for testing purposes
    # sql indexing is usually better than json indexing.
    def __init__(self, url, **kwargs):
        index = FieldsetInFilesWithJsonIndex.from_url(url)
        super().__init__(index, **kwargs)


source = IndexedUrlWithJsonIndex



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import inspect
import json
import re
from importlib import import_module

import requests
import yaml
from tqdm.auto import tqdm


def download_and_cache(url, return_none_on_404=False, **kwargs):
    from climetlab.sources.url import download_and_cache

    try:
        return download_and_cache(url, **kwargs)
    except requests.HTTPError as e:
        if return_none_on_404:
            if e.response is not None and e.response.status_code == 404:
                return None
        raise


def get_json(url: str, cache=False):
    if cache:
        with open(download_and_cache(url)) as f:
            return json.loads(f.read())

    r = requests.get(url)
    r.raise_for_status()
    return r.json()


def _dummy(**kwargs):
    pass


def consume_args(func1, func2, *args, **kwargs):
    # print("=====>", args, kwargs)

    if func1 is None:
        func1 = _dummy

    if func2 is None:
        func2 = _dummy

    args1 = set()
    sig1 = inspect.signature(func1)
    for name, param in sig1.parameters.items():
        if param.kind not in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            args1.add(name)

    # print(f"{func1.__name__}{sig1}")

    args2 = set()
    sig2 = inspect.signature(func2)
    for name, param in sig2.parameters.items():
        if param.kind not in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            args2.add(name)

    # print(f"{func2.__name__}{sig2}")

    common = args1.intersection(args2)
    common.discard("self")

    if common:
        raise NotImplementedError(
            f"{func1.__name__}{sig1} and {func2.__name__}{sig2} cannot share the same parameter(s): {common}"
        )

    spec = inspect.getfullargspec(func1)

    args_1 = []
    kwargs_1 = {}

    args = list(args)

    spec_args = [a for a in spec.args if a != "self"]

    for a in spec_args:
        if a in kwargs:
            break
        if not args:
            break
        args_1.append(args.pop(0))

    n = len(args_1)
    for a in spec_args[n:] + spec.kwonlyargs:
        if a in kwargs:
            kwargs_1[a] = kwargs.pop(a)

    # print('<=====', args_1, kwargs_1, args, kwargs)
    return args_1, kwargs_1, args, kwargs


def string_to_args(s):
    def typed(x):
        try:
            return int(x)
        except ValueError:
            pass

        try:
            return float(x)
        except ValueError:
            pass

        return x

    assert isinstance(s, str), s
    m = re.match(r"([\w\-]+)(\((.*)\))?", s)
    if not m:
        raise ValueError(f"Invalid argument '{s}'")

    name = m.group(1)

    if m.group(2) is None:
        return name, [], {}

    args = []
    kwargs = {}
    bits = [x.strip() for x in m.group(3).split(",") if x.strip()]
    for bit in bits:
        if "=" in bit:
            k, v = bit.split("=")
            kwargs[k.strip()] = typed(v.strip())
        else:
            args.append(typed(bit))

    return name, args, kwargs


def load_json_or_yaml(path):
    with open(path, "r") as f:
        if path.endswith(".json"):
            return json.load(f)
        if path.endswith(".yaml") or path.endswith(".yml"):
            return yaml.safe_load(f)
        raise ValueError(
            f"Cannot read file {path}. Need json or yaml with appropriate extension."
        )


def progress_bar(*, total=None, iterable=None, initial=0, desc=None):
    return tqdm(
        iterable=iterable,
        total=total,
        initial=initial,
        unit_scale=True,
        unit_divisor=1024,
        unit="B",
        disable=False,
        leave=False,
        desc=desc,
        # dynamic_ncols=True, # make this the default?
    )


MODULE_INSTALLED = {}


def module_installed(name):
    if name not in MODULE_INSTALLED:
        try:
            import_module(name)
            MODULE_INSTALLED[name] = True
        except ImportError:
            MODULE_INSTALLED[name] = False

    return MODULE_INSTALLED[name]


def module_loaded(name):
    from ..aaa import loaded_modules

    return name in loaded_modules()


class Separator:
    """
    >>> Separator.split("t+850")
    ['t', '850']
    >>> Separator.split(" t + 850 ")
    ['t', '850']
    >>> Separator.split(["t"])
    ['t']
    >>> Separator.join(["t", "850"])
    't+850'
    """

    SEPARATOR = "+"

    @classmethod
    def split(cls, arg):
        if arg is None:
            return None
        assert isinstance(arg, str), arg

        lst = []
        for n in arg.split(cls.SEPARATOR):
            n = n.strip()
            lst.append(n)
        return lst

    @classmethod
    def join(cls, iterable):
        assert not isinstance(iterable, str), iterable
        strings = [str(a) for a in iterable]
        if any(cls.SEPARATOR in str(a) for a in strings):
            raise ValueError(f"'{cls.SEPARATOR}' found in {strings}.")

        return cls.SEPARATOR.join(strings)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import os

from climetlab.exceptions import NotIndexedDirectoryError
from climetlab.readers.grib.index.sql import FieldsetInFilesWithSqlIndex
from climetlab.sources.indexed import IndexedSource

LOG = logging.getLogger(__name__)


def make_absolute(filename, root_dir, default):
    if filename is None:
        filename = default

    if os.path.isabs(filename):
        return filename

    absolute = os.path.join(root_dir, filename)
    LOG.debug(f"Transforming {filename} into absolute path {absolute}")
    return absolute


class GenericDirectorySource(IndexedSource):
    """Abstract class, INDEX_CLASS must be implemented"""

    INDEX_CLASS = None

    DEFAULT_JSON_FILE = "climetlab.index"
    DEFAULT_DB_FILE = "climetlab-2.db"

    def __init__(
        self,
        path,
        db_path=None,
        _index=None,
        **kwargs,
    ):
        """
        index_file: optional
            The input index location. Must be a JSON file.
            If None, parse the actual files in the directory pointed by "path".
        db_path: optional
            The actual used database file. Must be a SQL CliMetLab index file.
            If None, create one in the default location.

        If _index is not None, ignore all other arguments (for internal usage)
        """

        if _index is not None:  # for .sel()
            super().__init__(_index, **kwargs)
            return

        path = os.path.expanduser(path)

        self.path = path
        self.abspath = os.path.abspath(path)

        if db_path is None:
            db_path = make_absolute(
                db_path,
                self.abspath,
                default=self.DEFAULT_DB_FILE,
            )

        # Try to use db_path if it exists:
        if not os.path.exists(db_path):
            raise NotIndexedDirectoryError(
                f"This directory has not been indexed. Try running 'climetlab index_directory {self.path}'."
            )

        LOG.info(f"Using index file {db_path}")
        index = self.INDEX_CLASS.from_existing_db(db_path=db_path)
        super().__init__(index, **kwargs)


class IndexedDirectorySource(GenericDirectorySource):
    INDEX_CLASS = FieldsetInFilesWithSqlIndex


source = IndexedDirectorySource



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse
import cmd
import logging
import os
import sys
import traceback
from importlib import import_module

import entrypoints
from termcolor import colored

from .availability import AvailabilityCmd
from .benchmark import BenchmarkCmd
from .cache import CacheCmd
from .check import CheckCmd
from .completion import CompletionCmd
from .grib import GribCmd
from .grib_info import GribInfoCmd
from .settings import SettingsCmd
from .test_data import TestDataCmd

LOG = logging.getLogger(__name__)

try:
    import readline
except ImportError:  # Not availabe on win32

    class readline:
        def set_history_length(*args, **kwargs):
            pass

        def write_history_file(*args, **kwargs):
            pass

        def read_history_file(*args, **kwargs):
            pass


def get_plugins():
    plugins = []
    for e in entrypoints.get_group_all("climetlab.scripts"):
        module = import_module(e.module_name)
        klass = getattr(module, e.object_name)
        if klass in plugins:
            LOG.error(f"Potential plugins conflict for {module} {klass}.")
            continue
        plugins.append(klass)
    return plugins


class CliMetLabApp(
    cmd.Cmd,
    SettingsCmd,
    CompletionCmd,
    CacheCmd,
    CheckCmd,
    GribCmd,
    BenchmarkCmd,
    GribInfoCmd,
    AvailabilityCmd,
    TestDataCmd,
    *get_plugins(),
):
    # intro = 'Welcome to climetlab. Type ? to list commands.\n'
    prompt = colored("(climetlab) ", "yellow")

    rc_file = "~/.climetlab-history"

    def postloop(self):
        try:
            readline.set_history_length(1000)
            readline.write_history_file(os.path.expanduser(self.rc_file))
        except Exception:
            pass

    def preloop(self):
        try:
            readline.read_history_file(os.path.expanduser(self.rc_file))
            with open(os.path.expanduser(self.rc_file), "rt") as f:
                for line in f:
                    self.history.append(line[:-1])
        except Exception:
            pass

    def empty_line(self):
        pass

    def do_quit(self, args):
        """Quit climetlab."""
        return True

    def default(self, line):
        if line == "EOF":
            return True

        cmd = colored(line.split()[0], "yellow")
        help = colored("help", "yellow")

        print(
            f"Unknown command {cmd}. Type {help} for the list of known command names."
        )

    def onecmd(self, line):
        try:
            lst = line.split()

            def replace_dashes(txt):
                return txt.replace("-", "_")

            if len(lst) == 1:
                line = replace_dashes(lst[0])
            if len(lst) > 1:
                line = " ".join([replace_dashes(lst[0])] + lst[1:])
            return super().onecmd(line)
        except ValueError as e:
            traceback.print_exc()
            print(colored(str(e), "red"))
        except Exception:
            traceback.print_exc()
        return 33


def main():
    p = argparse.ArgumentParser(add_help=False)
    p.add_argument(
        "--debug",
        action="store_true",
        help="verbose operation (default: quiet)",
    )
    p.add_argument(
        "cmdline",
        type=str,
        metavar="CMD",
        nargs=argparse.REMAINDER,
    )

    p.add_argument(
        "-h", "--help", action="store_true", help="show this help message and exit"
    )
    p.add_argument("-v", "--version", action="store_true", help="show version and exit")
    args = p.parse_args()
    if args.version:
        from climetlab import __version__

        print(__version__)
        sys.exit()

    if args.help:
        p.print_help()
        cmdline = ["help"]
    else:
        cmdline = args.cmdline

    sys.argv[1:] = cmdline  # TODO: remove this?

    logging.basicConfig(level=args.debug and "DEBUG" or "WARN")

    app = CliMetLabApp()

    if cmdline:
        res = app.onecmd(" ".join(cmdline))
        if res:
            sys.exit(res)
    else:
        app.cmdloop()


def command_list():
    return [
        func[3:]
        for func in dir(CliMetLabApp)
        if callable(getattr(CliMetLabApp, func))
        and func.startswith("do_")
        and getattr(CliMetLabApp, func).__module__.startswith("climetlab.")
    ]


if __name__ == "__main__":
    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from climetlab.core.thread import SoftThreadPool
from climetlab.decorators import normalize
from climetlab.utils import tqdm

from .file import FileSource
from .prompt import APIKeyPrompt

LOG = logging.getLogger(__name__)


class MARSAPIKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://www.ecmwf.int/user/login/sso"
    retrieve_api_key_url = "https://api.ecmwf.int/v1/key/"

    prompts = [
        dict(
            name="url",
            default="https://api.ecmwf.int/v1",
            title="API url",
            validate=r"http.?://.*",
        ),
        dict(
            name="key",
            example="b295aad8af30332fad2fa8c963ab7900",
            title="API key",
            hidden=True,
            validate="[0-9a-z]{32}",
        ),
        dict(
            name="email",
            title="Your email",
        ),
    ]

    rcfile = "~/.ecmwfapirc"


class ECMWFApi(FileSource):
    def __init__(self, *args, **kwargs):
        super().__init__()

        request = {}
        for a in args:
            request.update(a)
        request.update(kwargs)

        requests = self.requests(**request)

        self.service()  # Trigger password prompt before threading

        nthreads = min(self.settings("number-of-download-threads"), len(requests))

        if nthreads < 2:
            self.path = [self._retrieve(r) for r in requests]
        else:
            with SoftThreadPool(nthreads=nthreads) as pool:
                futures = [pool.submit(self._retrieve, r) for r in requests]

                iterator = (f.result() for f in futures)
                self.path = list(tqdm(iterator, leave=True, total=len(requests)))

    def _retrieve(self, request):
        def retrieve(target, request):
            self.service().execute(request, target)

        return self.cache_file(
            retrieve,
            request,
        )

    @normalize("param", "variable-list(mars)")
    @normalize("date", "date-list(%Y-%m-%d)")
    @normalize("area", "bounding-box(list)")
    def requests(self, **kwargs):
        def value_to_list(v):
            if isinstance(v, (list, tuple)):
                return v
            return (v,)

        def to_lists(dic):
            # Ensure that we only give lists.
            # this is workaround because time=0 is treated as null or None and removed
            # and replaced by the default value with may be != 0 (such as 12)
            return {k: value_to_list(v) for k, v in dic.items()}

        split_on = kwargs.pop("split_on", None)
        if split_on is None or not isinstance(kwargs.get(split_on), (list, tuple)):
            kwargs = to_lists(kwargs)
            return [kwargs]

        result = []

        for v in kwargs[split_on]:
            r = dict(**kwargs)
            r[split_on] = v
            r = to_lists(r)
            result.append(r)

        return result

    def to_pandas(self, **kwargs):
        pandas_read_csv_kwargs = dict(
            sep="\t",
            comment="#",
            # parse_dates=["report_timestamp"],
            skip_blank_lines=True,
            skipinitialspace=True,
            compression="zip",
        )

        pandas_read_csv_kwargs.update(kwargs.get("pandas_read_csv_kwargs", {}))

        odc_read_odb_kwargs = dict(
            # TODO
        )

        odc_read_odb_kwargs.update(kwargs.get("odc_read_odb_kwargs", {}))

        return super().to_pandas(
            pandas_read_csv_kwargs=pandas_read_csv_kwargs,
            odc_read_odb_kwargs=odc_read_odb_kwargs,
            **kwargs,
        )



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import time

from climetlab import load_source
from climetlab.core.statistics import collect_statistics, retrieve_statistics

CML_BASEURL_S3 = "https://object-store.os-api.cci1.ecmwf.int/climetlab"
CML_BASEURL_CDS = "https://datastore.copernicus-climate.eu/climetlab"
CML_BASEURL_GET = "https://get.ecmwf.int/repository/test-data/climetlab"


def get_methods_list():
    methods = []
    # methods.append("cluster(2)|blocked(256)")
    # methods.append("cluster(5)|blocked(4096)")
    for i in [10, 100]:
        for j in [12, 16, 24]:
            methods.append(f"cluster({i})|blocked({2**j})")

    # for i in range(1,10,2):
    #     methods.append(f"sharp({10**i},1)")

    # for i in [1, 2, 3, 4, 5, 7, 10, 20, 50, 100, 500, 1000]:
    for i in [1, 5, 10, 50, 100]:
        methods.append(f"cluster({i})")

    methods.append("auto")

    for i in range(8, 25, 4):
        methods.append(f"blocked({2**i})")

    return methods


def check_grib_value(value, requested):
    if isinstance(requested, (list, tuple)):
        return any([check_grib_value(value, _v) for _v in requested])
    else:
        try:
            return int(value) == int(requested)
        except (TypeError, ValueError):
            return str(value) == str(requested)


def retrieve_and_check(index, request, range_method=None, **kwargs):
    print("--------")
    parts = index.get_path_offset_length(request)
    print("range_method", range_method)
    print("REQUEST", request)
    for url, p in parts:
        total = len(index.get_backend(url).entries)
        print(f"PARTS: {len(p)}/{total} parts in {url}")

    ####################
    # from climetlab import load_source
    # from climetlab.indexing import PerUrlIndex
    #
    # baseurl = "https://datastore.copernicus-climate.eu/climetlab"
    # s = load_source(
    #     "indexed-urls",
    #     PerUrlIndex(
    #         f"{baseurl}/test-data/input/indexed-urls/large_grib_1.grb",
    #     ),
    #     {"param": "r", "time": "1200"},
    #     range_method="auto",
    #     force=True,
    # )
    # assert 0, "Stop here"
    ####################

    now = time.time()
    s = load_source("indexed-urls", index, request, range_method=range_method, **kwargs)
    elapsed = time.time() - now
    print("ELAPSED", elapsed)
    try:
        paths = [s.path]
    except AttributeError:
        paths = [p.path for p in s.sources]

    for path in paths:
        # check that the downloaded gribs match the request
        for grib in load_source("file", path):
            for k, v in request.items():
                if k == "param":
                    k = "shortName"
                assert check_grib_value(grib._get(k), v), (grib._get(k), v)
    return elapsed


def radix(long, sep="("):
    assert long is not None, long
    if not isinstance(long, str):
        return long
    if sep not in long:
        return long
    return long.split(sep)[0]


def url_to_server(url):
    if url.startswith(CML_BASEURL_S3):
        return "EWC"
    if url.startswith(CML_BASEURL_CDS):
        return "CDS"
    if url.startswith(CML_BASEURL_GET):
        return "GET"
    return "Other"


def benchmark():
    collect_statistics(True)

    baseurls = [
        CML_BASEURL_S3,
        CML_BASEURL_CDS,
        # CML_BASEURL_GET,
    ]

    requests = [
        {"param": "r", "time": "1000", "step": "0"},
        {"param": "r", "time": "1000"},
        {"param": "r", "time": ["1100", "1200", "1300", "1400"]},
        {
            "param": ["r", "z"],
            "time": ["0200", "1000", "1800", "2300"],
            "levelist": ["500", "850"],
        },
        {"param": ["r", "z"], "levelist": ["500", "850"]},
        {"param": "r"},
        # {"param": ["r", "z"]},
        {"param": ["r", "z", "t"]},
        # {},
    ]

    methods = get_methods_list()

    # requests = [requests[2]]
    # methods = [methods[0]]
    # baseurls = [baseurls[0]]
    # requests = requests[::2]
    # methods = methods[::2]
    # baseurls = [baseurls[0]]
    failed = []
    successfull = 0
    import tqdm

    from climetlab.indexing import PerUrlIndex

    for request in tqdm.tqdm(requests):
        for range_method in tqdm.tqdm(methods):
            for baseurl in baseurls:
                index = PerUrlIndex(
                    f"{baseurl}/test-data/input/indexed-urls/large_grib_1.grb",
                )
                try:
                    retrieve_and_check(
                        index,
                        request,
                        range_method,
                        force=True,
                    )
                    successfull += 1
                except Exception as e:
                    failed.append((index, request, range_method))
                    print("FAILED for ", index, request, range_method)
                    print(e)

    stats = retrieve_statistics()

    run_id = get_run_id()

    logfiles = []

    path = f"climetlab_benchmark{run_id}.json"
    logfiles.append(path)
    stats.write_to_json(path)
    print(f"BENCHMARK FINISHED. Raw logs saved in {path}")

    df = stats.to_pandas()

    df["server"] = df["url"].apply(url_to_server)
    df["speed"] = df["total"] / df["elapsed"] / (1024 * 1024)  # MB/s
    df["method"] = df["full_method"].apply(radix)

    df = df.rename(
        dict(
            size_parts="size_requested",
            size_blocks="size_downloaded",
        )
    )
    df["size_ratio"] = df["size_downloaded"] / df["size_requested"]

    path = f"climetlab_benchmark{run_id}.csv"
    df.to_csv(path)
    # df.to_csv("climetlab_benchmark.csv")
    logfiles.append(path)

    print(f"Benchmark finished ({successfull} successfull, {len(failed)} failed).")
    print(
        "All data in the log files are anonymous."
        "Only the log file names contain personal data (machine name, IP, etc.)."
    )
    for f in logfiles:
        print(f"Log file: {f}")


def get_run_id(keys=("hostname", "ip", "date", "user", "time")):
    run_id = ""

    import datetime

    now = datetime.datetime.now()

    for k in keys:
        if k == "hostname":
            import socket

            run_id += "_" + str(socket.gethostname())
            continue

        if k == "user":
            import getpass

            run_id += "_" + str(getpass.getuser())
            continue

        if k == "ip":
            from requests import get

            ip = get("https://api.ipify.org").text
            run_id += "_" + str(ip)

        if k == "date":
            run_id += "_" + now.strftime("%Y-%m-%d")

        if k == "time":
            run_id += "_" + now.strftime("%H%M")

    return run_id



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime

from climetlab.wrappers import Wrapper


class DateWrapper(Wrapper):
    def __init__(self, data):
        self.data = data

    def to_datetime(self):
        return datetime.datetime(self.data.year, self.data.month, self.data.day)

    def to_datetime_list(self):
        return [self.to_datetime()]


def wrapper(data, *args, **kwargs):
    if isinstance(data, datetime.date):
        return DateWrapper(data)
    return None



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
import random

from tqdm import tqdm

import climetlab as cml

from .benchmarks.indexed_url import benchmark as benchmark_indexed_url
from .tools import experimental, parse_args

home = os.path.expanduser("~")

DATA_DATALOADING = {
    "grib": f"{home}/links/weather-bench-links/data-from-mc-symlinks-to-files-small/grib",
    "netcdf": f"{home}/links/weather-bench-links/data-from-ma/netcdf/pl_1999.nc",
    "b": f"{home}/dev/climetlab/tests/sources/gribs/b",
}


def benchmark_dataloading(*args):
    now = datetime.datetime.now
    import pandas as pd
    import xarray as xr

    class Exp:
        def __init__(self, directory):
            self.directory = directory

        def _get_selection(self):
            i_times = []
            dates = []
            n = int(args[2])  # number of dates to generate
            while len(i_times) < n:
                day = random.randint(1, 365)
                if day in i_times:
                    continue
                i_times.append(day)
                date = "1999" + pd.to_datetime(day, format="%j").strftime("%m%d")
                dates.append(date)
            assert len(dates) == n, (n, dates)
            print(dates)
            dates = sorted(dates)
            i_times = sorted(i_times)

            # dic = {"date": dates}
            dic = {"date": dates, "time": "0000"}
            # dic = {"date": dates, "time": "0000", "step": "0"}
            # dic = {"param":"z", "time": "0000", "step": "0"}
            # dic = {}
            # dic = {"date": "19990115"}
            # dic = {"date": ["19990115", "19990116"]}
            # dic = {"date": ["19990115", "19990116", "19990117", "19990118"]}
            return dic, dict(time=i_times)

    class GribExp(Exp):
        def get_ds(self):
            dic, isel_dic = self._get_selection()
            self.ds = cml.load_source("local", self.directory, dic)
            print(len(self.ds))

        def get_values(self):
            if args[1] == "direct":
                for one in tqdm(self.ds):
                    one.values.shape
                return

            elif args[1] == "xr":
                xds = self.ds.to_xarray()
                print(xds)
                for v in xds.variables:
                    print(v, xds[v].values.shape)
                return

            raise NotImplementedError()

    class NetcdfExp(Exp):
        def get_ds(self):
            self.ds = xr.open_dataset(self.directory)
            print(len(self.ds))

        def get_values(self):
            print(self.ds)

            dic, isel_dic = self._get_selection()
            ds = self.ds.isel(**isel_dic)

            for v in ds.variables:
                print(v, ds[v].values.shape)

    class ExpOnB(Exp):
        def _get_selection(self):
            return dict(param="2t", realization="0"), None

    def usage():
        print(
            """
        Usage:

        This script runs benchmark to get data, with two steps:
        1 - Get a handle on the dataset.
        2 - Then read the actual data.

        $0 grib direct 5:
        Read grib data with custom index.
        (Building climetlab.index json file if needed. Building .db in the database if needed.)
        Then read data directly asking eccode to decode the grib data.
        Using 5 dates.

        $0 grib xr 5:
        Same as above, but use to_xarray to read the data.
        This xarray calls cfgrib engine which calls climetlab which calls eccodes.

        $0 netcdf _ 5:
        Read the same data on netcdf files (second argument _ is a dummy argument) (TODO:yes, we need to use argparse).

        """
        )

    if len(args) == 0:
        usage()
        return

    d = args[0]
    d = DATA_DATALOADING.get(d, d)
    print(f"Using d={d}")

    if d.endswith(".nc") or d.endswith("netcdf"):
        exp = NetcdfExp(d)

    elif d.endswith("grib"):
        exp = GribExp(d)

    elif d.endswith("/b"):
        print("Todo fix this one")
        exp = ExpOnB(d)

    else:
        raise NotImplementedError()

    start = now()
    exp.get_ds()
    step_1 = now()
    exp.get_values()
    end = now()

    print("------------------")
    print("Time to get dataset handle:", step_1 - start)
    print("Time read the data from the dataset handle:", end - step_1)
    print("Total Time to actually read the data:", end - start)


class BenchmarkCmd:
    @parse_args(
        indexedurl=dict(
            action="store_true",
            help="Benchmark on using indexed URL (byte-range) and various servers.",
        ),
        dataloading=dict(
            action="store_true",
            help="Test loading some data.",
        ),
        nargs=dict(nargs="*"),
        all=dict(action="store_true", help="Run all benchmarks."),
    )
    @experimental
    def do_benchmark(self, args):
        """Run predefined benchmarks, for CliMetLab development purposes."""
        if args.all or args.indexedurl:
            print("Starting benchmark.")
            benchmark_indexed_url()

        if args.all or args.dataloading:
            print("Starting benchmark.")
            benchmark_dataloading(*args.nargs)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from urllib.parse import urlparse

import xarray as xr
import zarr

from . import Source

LOG = logging.getLogger(__name__)


class Cache:
    def __init__(self, store):
        self._store = store

    def __contains__(self, key):
        LOG.debug("S3 Cache.__contains__ %s", key)
        return key in self._store

    def __getitem__(self, key):
        LOG.debug("S3 Cache.__getitem__ %s", key)
        data = self._store[key]
        LOG.debug("S3 Cache.__getitem__ %d", len(data))
        return data

    def keys(self):
        return self._store.keys()


def url_to_s3_store(url, user=None, password=None):
    import s3fs

    bits = url.split("/")
    if bits[0] == "s3:":
        bits[0] = "https:"

    url = "/".join(bits[:3])
    root = "/".join(bits[3:])

    fs = s3fs.S3FileSystem(anon=True, client_kwargs={"endpoint_url": url})

    store = s3fs.S3Map(
        root=root,
        s3=fs,
        check=False,
    )

    store = Cache(store)

    store = zarr.storage.KVStore(store)

    return store


def find_store(store):
    o = urlparse(store)

    if "@" in o.netloc:
        auth, server = o.netloc.split("@")
        user, password = auth.split(":")

    if o.scheme in ["http", "https", "s3"]:
        return url_to_s3_store(store)
    if os.path.exists(store):
        if store.endswith(".zip"):
            return zarr.ZipStore(store)
        return store
    if o.scheme in ["file"]:
        # hard coded 3 because urlparse plays with the initial /
        return store[(len(o.scheme) + 3) :]

    raise NotImplementedError(f"Unknown protocol '{o.scheme}' for Zarr in {store}")


class Zarr(Source):
    def __init__(self, store, **kwargs):
        super().__init__(**kwargs)

        self._url = None

        if isinstance(store, str):
            self._url = store
            store = find_store(store)

        try:
            self._ds = xr.open_dataset(store, engine="zarr")  # TODO: chunks="auto" ?
        except zarr.errors.GroupNotFoundError as e:
            if self._url:
                LOG.error("ERROR : Cannot find data from %s", self._url)
            raise (e)
        except PermissionError as e:
            if self._url:
                LOG.error("ERROR : Permission denied on accessing %s", self._url)
            raise (e)
        except Exception as e:
            if self._url:
                LOG.error("ERROR when accessing %s", self._url)
            raise (e)

    def to_xarray(self):
        return self._ds


# this is already in Multi
# dsdict = {}
# for ds in dslist:
#     for value in ds[concat_dim].values:
#         dsdict[value] = ds.sel(**{concat_dim: value})
#     values_sorted = sorted(dsdict.keys())
#     dslist = [dsdict[d] for d in values_sorted]

# self._ds = xr.concat(dslist, dim=concat_dim)
# # self._ds = xr.open_mfdataset(stores, engine="zarr", combine='nested')


source = Zarr



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import metview as mv

from climetlab.core.ipython import ipython_active

if ipython_active:
    try:
        import ipywidgets  # noqa

        mv.setoutput("jupyter", plot_widget=True)
    except ImportError:
        mv.setoutput("jupyter", plot_widget=False)


def mv_read(*args, **kwargs):
    return mv.read(*args, **kwargs)


def mv_read_table(*args, **kwargs):
    return mv.read_table(*args, **kwargs)



#!/usr/bin/env python3
# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse

from climetlab.scripts.main import CliMetLabApp


def iprint(*args, indent=2):
    print(" " * indent, end="")
    print(*args)


def print_rst_underline(txt, padding="-", indent=0):
    iprint(txt, indent=indent)


def execute(*args):
    cmd_list = [(cmd, getattr(CliMetLabApp, cmd)) for cmd in dir(CliMetLabApp)]
    cmd_list = [(cmd, func) for (cmd, func) in cmd_list if cmd.startswith("do_")]
    cmd_list = [(cmd, func) for (cmd, func) in cmd_list if hasattr(func, "_argparser")]
    cmd_list = [(cmd[3:], func) for (cmd, func) in cmd_list]  # remove "do_"
    cmd_list = sorted(cmd_list, key=lambda x: x[0])

    assert len(cmd_list) > 0, "Documentation for climetlab CLI is empty"

    for cmd, func in cmd_list:
        module = func.__module__.split(".")[0]

        print()
        print("--------------------")
        print()
        print(f".. _{cmd}-command:")
        print()
        title = cmd
        if module != "climetlab":
            title = title + f"[{module}]"
        if getattr(func, "_climetlab_experimental", False):
            title = title + " (experimental)"
        print(title)
        print("-" * len(title))
        print()

        # Not garanteed to work with future versions of Python
        func._argparser.formatter_class = (
            lambda prog: argparse.RawDescriptionHelpFormatter(
                prog,
                width=90,
                max_help_position=10000,
            )
        )

        for n in func._argparser.format_help().split("\n"):
            if n.startswith("usage: "):
                n = "climetlab " + n[7:]
                print("Usage")
                print('"""""')
                iprint()
                iprint(".. code-block:: text")
                iprint()
                iprint(" ", n)
                continue

            if len(n.strip()) == 0:
                iprint()
                iprint()
                continue

            if n.endswith(" arguments:"):
                n = n[:-1]  # remove trailing ":"
                print(n[0].upper() + n[1:])
                print('"' * len(n))
                iprint()
                iprint(".. code-block:: text")
                iprint()
                continue

            if n.lower().startswith("examples:"):
                n = n[9:]
                print("Examples")
                print('""""""""')
                iprint()
                iprint(".. code-block:: text")
                iprint()
                iprint(" ", n)
                continue

            if n.strip().startswith("-"):
                iprint(" ", n)
                continue

            iprint(n)


if __name__ == "__main__":
    execute()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from ..datasets import load_dataset
from ..sources import load_source
from . import Source

LOG = logging.getLogger(__name__)


class LoadAction:
    def execute(self, v, data, last, inherit):
        if not isinstance(v, list):
            v = [v]
        for one in v:
            one = dict(**one)
            assert "name" in one, one
            name = one.pop("name")
            if inherit:
                last.update(one)
                one = last
            print(f"Using data from: {name}, {one}")
            source = self.load(name, **one)

            assert len(source), f"No data for {(name, one)}"
            data.append(source)


class LoadSource(LoadAction):
    def load(self, *args, **kwargs):
        return (load_source, args, kwargs)


class LoadDataset(LoadAction):
    def load(self, *args, **kwargs):
        return (load_dataset, args, kwargs)


class LoadConstants(LoadSource):
    def execute(self, v, data, last, inherit):
        super().execute(
            {
                "name": "constants",
                "source_or_dataset": data[0],
                "param": v,
            },
            data,
            last,
            inherit,
        )


ACTIONS = {
    "source": LoadSource,
    "dataset": LoadDataset,
    "constants": LoadConstants,
}


def instanciate_values(o, kwargs):
    if isinstance(o, dict):
        return {k: instanciate_values(v, kwargs) for k, v in o.items()}
    if isinstance(o, list):
        return [instanciate_values(v, kwargs) for v in o]
    if isinstance(o, tuple):
        return tuple([instanciate_values(v, kwargs) for v in o])
    if isinstance(o, str) and o.startswith("$"):
        return kwargs[o[1:]]
    return o


class Input:
    def __init__(self, config, **kwargs):
        from climetlab.utils import load_json_or_yaml

        if isinstance(config, str):
            config = load_json_or_yaml(config)
            if "input" in config:
                config = config["input"]
                config = instanciate_values(config, kwargs)

        assert isinstance(config, (list, tuple)), config

        self.config = config

    def expand(self):
        """
        The config provided to this "loader" source can have
        multiple sources/datasets. Let's iterate along each of
        them, and concatenate them as a unique source.
        """
        data = []
        inherit = False
        last = {}
        for input in self.config:
            assert isinstance(input, dict), input

            k = list(input.keys())[0]
            v = input[k]
            if k == "inherit":
                inherit = v
                assert len(input) == 1, input
                continue

            ACTIONS[k]().execute(v, data, last, inherit)

        return data


class Loader(Source):
    def __init__(self, config, **kwargs):
        self.input = Input(config, **kwargs)

    def mutate(self):
        sources = self.input.expand()
        source = sources[0]
        for s in sources[1:]:
            source += s
        from climetlab.readers.grib.index import GribFieldSet

        assert isinstance(source, GribFieldSet), type(source)

        return source


source = Loader



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools
from collections import defaultdict

import climetlab as cml
from climetlab import Source
from climetlab.core.temporary import temp_file
from climetlab.decorators import normalize
from climetlab.readers.grib.output import new_grib_output
from climetlab.utils.availability import Availability


class Accumulation:
    def __init__(self, out, param, date, time, step, number, stepping):
        self.out = out
        self.param = param
        self.date = date
        self.time = time * 100
        self.number = number
        self.steps = tuple(step)
        self.values = None
        self.seen = set()
        self.startStep = None
        self.endStep = None
        self.done = False
        self.stepping = stepping

    @property
    def key(self):
        return (self.param, self.date, self.time, self.steps, self.number)

    def add(self, field, values):
        step = field.metadata("step")
        if step not in self.steps:
            return

        assert not self.done, (self.key, step)

        startStep = field.metadata("startStep")
        endStep = field.metadata("endStep")

        assert endStep == step, (startStep, endStep, step)
        assert step not in self.seen, (self.key, step)

        assert endStep - startStep == self.stepping, (startStep, endStep)

        if self.startStep is None:
            self.startStep = startStep
        else:
            self.startStep = min(self.startStep, startStep)

        if self.endStep is None:
            self.endStep = endStep
        else:
            self.endStep = max(self.endStep, endStep)

        if self.values is None:
            import numpy as np

            self.values = np.zeros_like(values)

        self.values += values

        self.seen.add(step)

        if len(self.seen) == len(self.steps):
            self.out.write(
                self.values,
                template=field,
                startStep=self.startStep,
                endStep=self.endStep,
            )
            self.values = None
            self.done = True


class Era5Accumulations(Source):
    """This source use a MARS request to get the data for accumulated fields
    following the same logic as what has been done when storing ERA5 in the
    Climate Data Store (CDS).
    The date+time provided to this source refers to the valid datetime of the data,
    these date+time are used to compute the date+time+step that are actually used to perform the mars request.

    Note 1 : as all these are lists, there are potentially to many fields requested to MARS.
          This is taken care by the final ".sel" .
    Note 2 : There are no overlap due to the way the date+time+step are computed
    """

    def __init__(self, *args, **kwargs):
        request = {}
        for a in args:
            request.update(self.requests(**a))
        request = self.requests(**kwargs)

        param = request["param"]
        if not isinstance(param, (list, tuple)):
            param = [param]

        for p in param:
            assert p in ["cp", "lsp", "tp"], p

        number = request.get("number", [0])
        assert isinstance(number, (list, tuple))

        user_step = 6  # For now, we only support 6h accumulation

        user_dates = request["date"]
        user_times = request["time"]

        requested = set()

        era_request = dict(**request)

        type_ = request.get("type", "an")
        if type_ == "an":
            type_ = "fc"

        stepping = 1
        if request.get("stream") == "enda":
            stepping = 3
            for n in user_times:
                assert n % 6 == 0, n

        era_request.update({"class": "ea", "type": type_, "levtype": "sfc"})

        tmp = temp_file()
        path = tmp.path
        out = new_grib_output(path)

        requests = []

        for user_date, user_time in itertools.product(user_dates, user_times):
            assert isinstance(user_date, datetime.datetime), (
                type(user_date),
                user_dates,
                user_times,
            )
            assert isinstance(user_time, int), (type(user_time), user_dates, user_times)
            assert 0 <= user_time <= 24, user_time

            requested.add(user_date + datetime.timedelta(hours=user_time))

            when = (
                user_date
                + datetime.timedelta(hours=user_time)
                - datetime.timedelta(hours=user_step)
            )
            add_step = 0

            while when.hour not in (6, 18):
                when -= datetime.timedelta(hours=stepping)
                add_step += stepping

            steps = tuple(
                step + add_step
                for step in range(stepping, user_step + stepping, stepping)
            )

            for p in param:
                for n in number:
                    requests.append(
                        {
                            "param": p,
                            "date": int(when.strftime("%Y%m%d")),
                            "time": when.hour,
                            "step": sorted(steps),
                            "number": n,
                        }
                    )

        compressed = Availability(requests)
        ds = cml.load_source("empty")
        for r in compressed.iterate():
            era_request.update(r)
            ds = ds + cml.load_source("mars", **era_request)

        accumulations = defaultdict(list)
        for a in [Accumulation(out, stepping=stepping, **r) for r in requests]:
            for s in a.steps:
                accumulations[(a.param, a.date, a.time, s, a.number)].append(a)

        for field in ds:
            key = (
                field.metadata("param"),
                field.metadata("date"),
                field.metadata("time"),
                field.metadata("step"),
                field.metadata("number"),
            )
            values = field.values  # optimisation
            for a in accumulations[key]:
                a.add(field, values)

        for acc in accumulations.values():
            for a in acc:
                assert a.done, (a.key, a.seen, a.steps)

        out.close()

        ds = cml.load_source("file", path)

        self.ds = cml.load_source("file", path)
        assert len(self.ds) / len(param) / len(number) == len(requested), (
            len(self.ds),
            len(param),
            len(requested),
        )
        self.ds._tmp = tmp

    def mutate(self):
        return self.ds

    @normalize("date", "date-list(datetime.datetime)")
    @normalize("time", "int-list")
    @normalize("area", "bounding-box(list)")
    def requests(self, **kwargs):
        result = dict(**kwargs)

        return result


source = Era5Accumulations



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from .. import Reader
from .fieldset import NetCDFFieldSetFromFile


class NetCDFReader(Reader, NetCDFFieldSetFromFile):
    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        NetCDFFieldSetFromFile.__init__(self, path)


def reader(source, path, magic=None, deeper_check=False):
    if magic is None or magic[:4] in (b"\x89HDF", b"CDF\x01", b"CDF\x02"):
        return NetCDFReader(source, path)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from xml.dom import minidom

from climetlab.utils import download_and_cache

URL = "https://cfconventions.org/Data/cf-standard-names/{version}/src/cf-standard-name-table.xml"


def cf_standard_names(version=78):
    path = download_and_cache(URL.format(version=version))
    xmldoc = minidom.parse(path)
    entries = xmldoc.getElementsByTagName("entry")
    for entry in entries:
        yield entry.attributes["id"].value


if __name__ == "__main__":
    for n in cf_standard_names():
        print(n)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json

from climetlab.utils import download_and_cache

URL = "https://apps.ecmwf.int/codes/grib/json/"

PARAMS = {}


def _param_id_dict():
    if not PARAMS:
        path = download_and_cache(URL)
        with open(path) as f:
            entries = json.load(f)

        for entry in entries["parameters"]:
            PARAMS[int(entry["param_id"])] = entry

    return PARAMS


def param_id_to_dict(param_id):
    dic = _param_id_dict()
    return dic[param_id]


def param_id_to_short_name(param_id):
    entry = param_id_to_dict(param_id)
    return entry["param_shortName"]


if __name__ == "__main__":
    import sys

    for i in sys.argv[1:]:
        print(param_id_to_short_name(int(i)))



#!/usr/bin/env python
#
# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#
import csv
import logging
import os

from climetlab.utils import load_json_or_yaml

LOG = logging.getLogger(__name__)

ALIASES = {}


def load_csv(path):
    result = {}
    with open(path) as f:
        for row in csv.reader(f):
            result[row[0]] = row[1]

    return result


def _find_aliases(name):
    if name not in ALIASES:
        path = os.path.join(os.path.dirname(__file__), name)
        if os.path.exists(path + ".csv"):
            ALIASES[name] = load_csv(path + ".csv")
        else:
            ALIASES[name] = load_json_or_yaml(path + ".yaml")
    return ALIASES[name]


def unalias(name, value):
    return _find_aliases(name).get(value, value)


if __name__ == "__main__":
    print(unalias("grib-paramid", "2t"))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)


class PandasMixIn:
    def to_pandas(self, latitude=None, longitude=None, **kwargs):
        import pandas as pd

        def ident(x):
            return x

        filter = ident

        def select_point(d):
            return [x for x in d if x["lat"] == latitude and x["lon"] == longitude]

        if latitude is not None or longitude is not None:
            filter = select_point

        frames = []
        for s in self:
            df = pd.DataFrame(filter(s.data))
            df["datetime"] = s.valid_datetime()
            for k, v in s.as_mars().items():
                df[k] = v
            frames.append(df)
        return pd.concat(frames)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from .tools import parse_args


class TestDataCmd:
    @parse_args(
        directory=(
            None,
            dict(
                metavar="DIRECTORY",
                help="Shell to use for autocompletion. Must be zsh or bash.",
                nargs="?",
            ),
        ),
    )
    def do_test_data(self, args):
        """
        Create a directory with data used to test climetlab.

        """
        from climetlab.testing import build_testdata

        directory = args.directory
        if not directory:
            directory = "./test-data"

        print(f"Adding testdata in {directory}")
        build_testdata(directory)
        print(f"Added testdata in {directory}")



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# The code is copied from skinnywms, and we should combile later

import datetime


def as_datetime(self, time):
    return datetime.datetime.strptime(str(time)[:19], "%Y-%m-%dT%H:%M:%S")


def as_level(self, level):
    n = float(level)
    if int(n) == n:
        return int(n)
    return n


class Slice:
    def __init__(self, name, value, index, is_dimension, is_info):
        self.name = name
        self.index = index
        self.value = value
        self.is_dimension = (not is_info,)
        self.is_info = is_info

    def __repr__(self):
        return "[%s:%s=%s]" % (self.name, self.index, self.value)


class TimeSlice(Slice):
    pass


class LevelSlice(Slice):
    pass


class Coordinate:
    def __init__(self, variable, info):
        self.variable = variable
        # We only support 1D coordinate for now
        # assert len(variable.dims) == 1
        self.is_info = info
        self.is_dimension = not info

        if variable.values.ndim == 0:
            self.values = [self.convert(variable.values)]
        else:
            self.values = [self.convert(t) for t in variable.values.flatten()]

    def make_slice(self, value):
        return self.slice_class(
            self.variable.name,
            value,
            self.values.index(value),
            self.is_dimension,
            self.is_info,
        )

    def __repr__(self):
        return "%s[name=%s,values=%s]" % (
            self.__class__.__name__,
            self.variable.name,
            len(self.values),
        )


class TimeCoordinate(Coordinate):
    slice_class = TimeSlice
    is_dimension = True
    convert = as_datetime


class LevelCoordinate(Coordinate):
    # This class is just in case we want to specialise
    # 'level', othewise, it is the same as OtherCoordinate
    slice_class = LevelSlice
    is_dimension = False
    convert = as_level


class OtherCoordinate(Coordinate):
    slice_class = Slice
    is_dimension = False
    convert = as_level



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import re
import sys

from .tools import parse_args

LOG = logging.getLogger(__name__)


class GribInfoCmd:
    @parse_args(
        param=dict(type=str, help="Comma separated list of parameters."),
        json=dict(action="store_true", help="Long json output format."),
        stdin=dict(action="store_true", help="Using stdin as input (for bash piping)."),
    )
    def do_grib_info(self, args):
        """
        Display information about grib parameters.
        """
        from climetlab.vocabularies.grib import param_id_to_dict

        def iter_params():
            if args.param:
                for p in args.param.split(","):
                    yield p
            if args.stdin:
                for line in sys.stdin:
                    line = line.strip("\n").strip("\r\n")
                    for lst in re.split(r"\s+", line):
                        for lst_ in re.split(",", lst):
                            yield lst_

        for param in iter_params():
            entry = None

            if param.isdigit():
                p = int(param)
                entry = param_id_to_dict(p)

            if entry is None:
                raise NotImplementedError(f"No info found for {param}")

            if args.json:
                s = json.dumps(entry, sort_keys=True, indent=4)
                print(s)
            else:
                print(entry["param_shortName"])



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from . import Reader

LOG = logging.getLogger(__name__)


class Unknown(Reader):
    def __init__(self, source, path, magic):
        super().__init__(source, path)
        self.magic = magic
        LOG.warning("Unknown file type %s (%s), ignoring", path, magic)

    def ignore(self):
        # Used by multi-source
        return True

    def __len__(self):
        return 0



# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


class Table:
    def __init__(self, rows, columns):
        self.rows = rows
        self.columns = columns
        self.elements = {}

    def __setitem__(self, key, value):
        self.elements[key] = value

    def __getitem__(self, key):
        return self.elements[key]

    def _repr_html_(self):
        result = []
        result.append('<table style="padding:0;margin:0;">')
        for r in range(self.rows):
            result.append('<tr style="padding:0;margin:0;">')
            for c in range(self.columns):
                result.append('<td style="padding:0;margin:0;">')
                if (r, c) in self.elements:
                    result.append(self.render(self.elements[(r, c)]))
                else:
                    result.append("&nbsp;")
                result.append("</td>")
            result.append("</tr>")
        result.append("</table>")
        return "\n".join(str(s) for s in result)

    def render(self, element):
        e = element.render()
        src, attrs = e._repr_png_()
        return f'<img src="data:image/png;base64, {src}">'



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import copy
import logging

LOG = logging.getLogger(__name__)


def deep_update(old, new):
    # deep update, merging dictionaries
    assert isinstance(new, dict), f"Expecting a dict, but received: {new}"
    for k, v in new.items():
        v = new[k]
        if k in old and isinstance(old[k], dict):
            deep_update(old[k], v)
        else:
            old[k] = v
    return old


class Kwargs(dict):
    def __init__(
        self,
        user,
        default=None,
        forced=None,
        logging_owner="",
        logging_main_key="",
    ):
        if default is None:
            default = {}

        if forced is None:
            forced = {}

        kwargs = copy.deepcopy(default)

        for k, v in user.items():
            if k in forced and v != forced[k]:
                LOG.warning(
                    (
                        f"In {logging_owner} {logging_main_key},"
                        f"ignoring attempt to override {k}={forced[k]} with {k}={v}."
                    )
                )
                continue

            if k in default and v != default[k]:
                LOG.warning(
                    (
                        f"In {logging_owner} {logging_main_key}, overriding the default value "
                        f"({k}={default[k]}) with {k}={v} is not recommended."
                    )
                )

            kwargs[k] = v

        kwargs.update(forced)

        super().__init__(kwargs)


# TODO: add test (and fix bugs)
def merge_dicts(*dicts):
    result = {}
    for v in dicts:
        result = deep_update(result, v)
    return result



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import re

import yaml

from climetlab.core.data import get_data_entry
from climetlab.utils.kwargs import merge_dicts

LOG = logging.getLogger(__name__)

CURRENT_DEPLOYS = []


def local():
    from dask.distributed import LocalCluster

    return LocalCluster


def ssh():
    from dask.distributed import SSHCluster

    return SSHCluster


def slurm():
    from dask_jobqueue import SLURMCluster

    return SLURMCluster


CLUSTERS = {
    "ssh": ssh,
    "SSHCluster": ssh,
    "local": local,
    "LocalCluster": local,
    "slurm": slurm,
    "SLURMCluster": slurm,
}


def resolve_cluster_name(name):
    return CLUSTERS[name]()


class DaskDeploy:
    _scale = None
    client = None

    def __init__(
        self,
        start_client=True,
        cluster_cls=None,
        cluster_kwargs=None,
        client_kwargs=None,
        scale=None,
    ):
        from dask.distributed import Client

        if cluster_kwargs is None:
            cluster_kwargs = {}
        if client_kwargs is None:
            client_kwargs = {}

        self.cluster = None
        self.client = None
        self._scale = scale

        self.cluster_class = resolve_cluster_name(cluster_cls)
        self.cluster_kwargs = cluster_kwargs
        LOG.debug(f"Creating Dask cluster: {self.cluster_class}({cluster_kwargs})")
        self.cluster = self.cluster_class(**self.cluster_kwargs)
        LOG.debug(f"Dask cluster={self.cluster}")

        self.scale()

        if start_client:
            self.client_kwargs = client_kwargs
            LOG.debug(f"Starting client: {client_kwargs}")
            self.client = Client(self.cluster, **self.client_kwargs)
            LOG.debug(f"Dask client={self.client}")

        CURRENT_DEPLOYS.append(self)

    def scale(self):
        if not self._scale:
            return
        print(f"Scaling to {self._scale}")
        self.cluster.scale(self._scale)

    def shutdown(self):
        if self.client:
            self.client.close()
        self.cluster.close()

    def __str__(self):
        return f"Cluster={self.cluster}, Client={self.client}"

    def _repr_html_(self):
        dashboard = self.cluster.dashboard_link
        dashboard = re.sub(
            "http://[0-9\\.]*:", "https://localhost:48167/proxy/", dashboard
        )
        return f"Cluster={self.cluster}, Client={self.client}, Dashboard:<a href='{dashboard}'>{dashboard}</a>"

    @property
    def dashboard_link(self):
        # TODO: make this work everywhere
        return self.cluster.dashboard_link


def start_dask(name_or_yaml_filename, **kwargs):
    if len(CURRENT_DEPLOYS) > 0:
        LOG.warn(
            f"Creating multiple dask clusters ({len(CURRENT_DEPLOYS)}) already running)."
        )

    _, ext = os.path.splitext(name_or_yaml_filename)
    if ext in (".yaml", ".yml"):
        # Explicit yaml file is provided
        filename = os.path.expanduser(name_or_yaml_filename)
        LOG.debug(f"Using yaml file {filename} to configure dask.")
        with open(filename) as f:
            config = yaml.load(f, Loader=yaml.SafeLoader)
    else:
        # The name (name_or_yaml_filename) refers to a yaml file in the climetlab//data
        # or user config ($HOME/.climetlab/dask).
        config = get_data_entry("dask", name=name_or_yaml_filename).data

    if "dask" in config:
        config = config["dask"]

    # kwargs always overwrite the config file values.
    config = merge_dicts(config, kwargs)

    LOG.debug(f"Creating a dask deployment with dask config {config}.")
    return DaskDeploy(**config)


def performance_report(*args, **kwargs):
    import dask.distributed

    return dask.distributed.performance_report(*args, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import functools
import getpass
import logging
import os
import tempfile
from contextlib import contextmanager
from typing import Callable

import yaml

from climetlab.utils.html import css
from climetlab.utils.humanize import as_bytes, as_percent, as_seconds
from climetlab.version import __version__ as VERSION

LOG = logging.getLogger(__name__)

DOT_CLIMETLAB = os.path.expanduser("~/.climetlab")


class Setting:
    def __init__(self, default, description, getter=None, none_ok=False, kind=None):
        self.default = default
        self.description = description
        self.getter = getter
        self.none_ok = none_ok
        self.kind = kind if kind is not None else type(default)

    def kind(self):
        return type(self.default)

    def save(self, name, value, f):
        for n in self.description.split("\n"):
            print(f"# {n.strip()}", file=f)
        print(file=f)
        comment = yaml.dump({name: self.default}, default_flow_style=False)
        for n in comment.split("\n"):
            if n:
                print(f"# {n}", file=f)
        if value != self.default:
            print(file=f)
            yaml.dump({name: value}, f, default_flow_style=False)


_ = Setting


SETTINGS_AND_HELP = {
    "cache-directory": _(
        os.path.join(tempfile.gettempdir(), "climetlab-%s" % (getpass.getuser(),)),
        """Directory of where the downloaded files are cached, with ``${USER}`` is the user id.
        See :doc:`/guide/caching` for more information.""",
    ),
    "styles-directories": _(
        [os.path.join(DOT_CLIMETLAB, "styles")],
        """List of directories where to search for styles definitions.
        See :ref:`styles` for more information.""",
    ),
    "projections-directories": _(
        [os.path.join(DOT_CLIMETLAB, "projections")],
        """List of directories where to search for projections definitions.
        See :ref:`projections` for more information.""",
    ),
    "layers-directories": _(
        [os.path.join(DOT_CLIMETLAB, "layers")],
        """List of directories where to search for layers definitions.
        See :ref:`layers` for more information.""",
    ),
    "dask-directories": _(
        [os.path.join(DOT_CLIMETLAB, "dask")],
        """List of directories where to search for dask cluster definitions.
        See :ref:`dask` for more information.""",
    ),
    "datasets-directories": _(
        [os.path.join(DOT_CLIMETLAB, "datasets")],
        """List of directories where to search for datasets definitions.
        See :ref:`datasets` for more information.""",
    ),
    "datasets-catalogs-urls": _(
        ["https://github.com/ecmwf-lab/climetlab-datasets/raw/main/datasets"],
        """List of url where to search for catalogues of datasets definitions.
        See :ref:`datasets` for more information.""",
    ),
    "plotting-options": _(
        {},
        """Dictionary of default plotting options.
        See :ref:`plotting` for more information.""",
    ),
    "number-of-download-threads": _(
        5,
        """Number of threads used to download data.""",
    ),
    "maximum-cache-size": _(
        None,
        """Maximum disk space used by the CliMetLab cache (ex: 100G or 2T).""",
        getter="_as_bytes",
        none_ok=True,
    ),
    "maximum-cache-disk-usage": _(
        "90%",
        """Disk usage threshold after which CliMetLab expires older cached entries (% of the full disk capacity).
        See :doc:`/guide/caching` for more information.""",
        getter="_as_percent",
    ),
    "url-download-timeout": _(
        "30s",
        """Timeout when downloading from an url.""",
        getter="_as_seconds",
    ),
    "check-out-of-date-urls": _(
        True,
        "Perform a HTTP request to check if the remote version of a cache file has changed",
    ),
    "download-out-of-date-urls": _(
        False,
        "Re-download URLs when the remote version of a cached file as been changed",
    ),
    "use-standalone-mars-client-when-available": _(
        True,
        "Use the standalone mars client when available instead of using the web API.",
    ),
}


NONE = object()
DEFAULTS = {}
for k, v in SETTINGS_AND_HELP.items():
    DEFAULTS[k] = v.default


@contextmanager
def new_settings(s):
    SETTINGS._stack.append(s)
    try:
        yield None
    finally:
        SETTINGS._stack.pop()
        SETTINGS._notify()


def forward(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        if self._stack:
            return func(self._stack[-1], *args, **kwargs)
        return func(self, *args, **kwargs)

    return wrapped


def save_settings(path, settings):
    LOG.debug("Saving settings")
    with open(path, "w") as f:
        print("# This file is automatically generated", file=f)
        print(file=f)

        for k, v in sorted(settings.items()):
            h = SETTINGS_AND_HELP.get(k)
            if h:
                print(file=f)
                print("#", "-" * 76, file=f)
                h.save(k, v, f)

        print(file=f)
        print("#", "-" * 76, file=f)
        print("# Version of CliMetLab", file=f)
        print(file=f)
        yaml.dump({"version": VERSION}, f, default_flow_style=False)
        print(file=f)


class Settings:
    def __init__(self, settings_yaml: str, defaults: dict, callbacks=[]):
        self._defaults = defaults
        self._settings = dict(**defaults)
        self._callbacks = [c for c in callbacks]
        self._settings_yaml = settings_yaml
        self._pytest = None
        self._stack = []

    @forward
    def get(self, name: str, default=NONE):
        """[summary]

        Args:
            name (str): [description]
            default ([type], optional): [description]. Defaults to NONE.

        Returns:
            [type]: [description]
        """

        if name not in SETTINGS_AND_HELP:
            raise KeyError("No setting name '%s'" % (name,))

        getter, none_ok = (
            SETTINGS_AND_HELP[name].getter,
            SETTINGS_AND_HELP[name].none_ok,
        )
        if getter is None:
            getter = lambda name, value, none_ok: value  # noqa: E731
        else:
            getter = getattr(self, getter)

        if default is NONE:
            return getter(name, self._settings[name], none_ok)

        return getter(name, self._settings.get(name, default), none_ok)

    @forward
    def set(self, name: str, *args, **kwargs):
        """[summary]

        Args:
            name (set): [description]
            value ([type]): [description]
        """

        if name not in SETTINGS_AND_HELP:
            raise KeyError("No setting name '%s'" % (name,))

        klass = SETTINGS_AND_HELP[name].kind

        if klass in (bool, int, float, str):
            # TODO: Proper exceptions
            assert len(args) == 1
            assert len(kwargs) == 0
            value = args[0]
            value = klass(value)

        if klass is list:
            assert len(args) > 0
            assert len(kwargs) == 0
            value = list(args)
            if len(args) == 1 and isinstance(args[0], list):
                value = args[0]

        if klass is dict:
            assert len(args) <= 1
            if len(args) == 0:
                assert len(kwargs) > 0
                value = kwargs

            if len(args) == 1:
                assert len(kwargs) == 0
                value = args[0]

        getter, none_ok = (
            SETTINGS_AND_HELP[name].getter,
            SETTINGS_AND_HELP[name].none_ok,
        )
        if getter is not None:
            assert len(args) == 1
            assert len(kwargs) == 0
            value = args[0]
            # Check if value is properly formatted for getter
            getattr(self, getter)(name, value, none_ok)
        else:
            if not isinstance(value, klass):
                raise TypeError("Setting '%s' must be of type '%s'" % (name, klass))

        self._settings[name] = value
        self._changed()

    @forward
    def reset(self, name: str = None):
        """Reset setting(s) to default values.

        Args:
            name (str, optional): The name of the setting to reset to default. If the setting does not have a default,
            it is removed. If `None` is passed, all settings are reset to their default values. Defaults to None.
        """
        if name is None:
            self._settings = dict(**DEFAULTS)
        else:
            if name not in DEFAULTS:
                raise KeyError("No setting name '%s'" % (name,))

            self._settings.pop(name, None)
            if name in DEFAULTS:
                self._settings[name] = DEFAULTS[name]
        self._changed()

    @forward
    def _repr_html_(self):
        html = [css("table")]
        html.append("<table class='climetlab'>")
        for k, v in sorted(self._settings.items()):
            setting = SETTINGS_AND_HELP.get(k, None)
            default = setting.default if setting else ""
            html.append("<tr><td>%s</td><td>%r</td><td>%r</td></td>" % (k, v, default))
        html.append("</table>")
        return "".join(html)

    @forward
    def dump(self):
        for k, v in sorted(self._settings.items()):
            yield ((k, v, SETTINGS_AND_HELP.get(k)))

    def _changed(self):
        self._save()
        self._notify()

    def _notify(self):
        for cb in self._callbacks:
            cb()

    def on_change(self, callback: Callable[[], None]):
        self._callbacks.append(callback)

    def _save(self):
        if self._settings_yaml is None:
            return

        try:
            save_settings(self._settings_yaml, self._settings)
        except Exception:
            LOG.error(
                "Cannot save CliMetLab settings (%s)",
                self._settings_yaml,
                exc_info=True,
            )

    def _as_bytes(self, name, value, none_ok):
        return as_bytes(value, name=name, none_ok=none_ok)

    def _as_percent(self, name, value, none_ok):
        return as_percent(value, name=name, none_ok=none_ok)

    def _as_seconds(self, name, value, none_ok):
        return as_seconds(value, name=name, none_ok=none_ok)

    # def _as_number(self, name, value, units, none_ok):
    #     return as_number(name, value, units, none_ok)

    @forward
    def temporary(self, name=None, *args, **kwargs):
        tmp = Settings(None, self._settings, self._callbacks)
        if name is not None:
            tmp.set(name, *args, **kwargs)
        return new_settings(tmp)


save = False
settings_yaml = os.path.expanduser("~/.climetlab/settings.yaml")

try:
    if not os.path.exists(DOT_CLIMETLAB):
        os.mkdir(DOT_CLIMETLAB, 0o700)
    if not os.path.exists(settings_yaml):
        save_settings(settings_yaml, DEFAULTS)
except Exception:
    LOG.error(
        "Cannot create CliMetLab settings directory, using defaults (%s)",
        settings_yaml,
        exc_info=True,
    )

settings = dict(**DEFAULTS)
try:
    with open(settings_yaml) as f:
        s = yaml.load(f, Loader=yaml.SafeLoader)
        if not isinstance(s, dict):
            s = {}

        settings.update(s)

    # if s != settings:
    #     save = True

    def parse(v):
        parts = v.split(".")
        major = int(parts[0])
        minor = int(parts[1]) if len(parts) > 1 else 0
        return major, minor

    if parse(settings.get("version")) < parse(VERSION):
        save = True

except Exception:
    LOG.error(
        "Cannot load CliMetLab settings (%s), reverting to defaults",
        settings_yaml,
    )
    save = True

SETTINGS = Settings(settings_yaml, settings)
if save:
    SETTINGS._save()



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
from collections import defaultdict

from climetlab.utils import download_and_cache


class Part:
    def __init__(self, path, offset, length):
        assert path is not None
        self.path = path
        self.offset = offset
        self.length = length

    def __eq__(self, other):
        return (
            self.path == other.path
            and self.offset == other.offset
            and self.length == other.length
        )

    @classmethod
    def resolve(cls, parts, directory=None):
        paths = defaultdict(list)
        for i, part in enumerate(parts):
            paths[part.path].append(part)

        for path, bits in paths.items():
            if (
                path.startswith("http://")
                or path.startswith("https://")
                or path.startswith("ftp://")
            ):
                newpath = download_and_cache(
                    path, parts=[(p.offset, p.length) for p in bits]
                )
                newoffset = 0
                for p in bits:
                    p.path = newpath
                    p.offset = newoffset
                    newoffset += p.length

            elif directory and not os.path.isabs(path):
                for p in bits:
                    p.path = os.path.join(directory, path)

        return parts

    def __repr__(self):
        return f"Part[{self.path},{self.offset},{self.length}]"



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

LOG = logging.getLogger(__name__)


def reader(source, path, magic=None, deeper_check=False):
    if magic is None or magic[:4] == b"GRIB":
        from .reader import GRIBReader

        return GRIBReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import functools
import logging
import math
from abc import abstractmethod
from collections import defaultdict

import climetlab as cml
from climetlab.core.order import build_remapping, normalize_order_by
from climetlab.core.select import normalize_selection
from climetlab.sources import Source

LOG = logging.getLogger(__name__)


class OrderOrSelection:
    def __str__(self):
        return f"{self.__class__.__name__}({self.kwargs})"

    @property
    def is_empty(self):
        return not self.kwargs


class Selection(OrderOrSelection):
    def __init__(self, kwargs, remapping=None):
        self.remapping = build_remapping(remapping)

        class InList:
            def __init__(self, lst):
                self.first = True
                self.lst = lst  # lazy casting: lst will be modified

            def __call__(self, x):
                if self.first and x is not None:
                    cast = type(x)
                    self.lst = [cast(y) for y in self.lst]
                    self.first = False
                return x in self.lst

        self.actions = {}
        for k, v in kwargs.items():
            if v is None or v is cml.ALL:
                self.actions[k] = lambda x: True
                continue

            if callable(v):
                self.actions[k] = v
                continue

            if not isinstance(v, (list, tuple, set)):
                v = [v]

            v = set(v)

            self.actions[k] = InList(v)

    def match_element(self, element):
        metadata = self.remapping(element.metadata)
        return all(v(metadata(k)) for k, v in self.actions.items())


class OrderBase(OrderOrSelection):
    def __init__(self, kwargs, remapping):
        self.actions = self.build_actions(kwargs)
        self.remapping = remapping

    @abstractmethod
    def build_actions(self, kwargs):
        raise NotImplementedError()

    def compare_elements(self, a, b):
        assert callable(self.remapping), (type(self.remapping), self.remapping)
        a_metadata = self.remapping(a.metadata)
        b_metadata = self.remapping(b.metadata)
        for k, v in self.actions.items():
            n = v(a_metadata(k), b_metadata(k))
            if n != 0:
                return n
        return 0


class Order(OrderBase):
    def build_actions(self, kwargs):
        actions = {}

        def ascending(a, b):
            if a == b:
                return 0
            if a > b:
                return 1
            if a < b:
                return -1
            raise ValueError(f"{a},{b}")

        def descending(a, b):
            if a == b:
                return 0
            if a > b:
                return -1
            if a < b:
                return 1
            raise ValueError(f"{a},{b}")

        class Compare:
            def __init__(self, order):
                self.order = order

            def __call__(self, a, b):
                return ascending(self.get(a), self.get(b))

            def get(self, x):
                return self.order[x]

        for k, v in kwargs.items():
            if v == "ascending" or v is None:
                actions[k] = ascending
                continue

            if v == "descending":
                actions[k] = descending
                continue

            if callable(v):
                actions[k] = v
                continue

            assert isinstance(
                v, (list, tuple)
            ), f"Invalid argument for {k}: {v} ({type(v)})"

            order = {}
            for i, key in enumerate(v):
                order[str(key)] = i
                try:
                    order[int(key)] = i
                except ValueError:
                    pass
                except TypeError:
                    print('Cannot convert "%s" to int (%s)' % (key, type(key)))
                    raise
                try:
                    order[float(key)] = i
                except ValueError:
                    pass
            actions[k] = Compare(order)

        return actions


class Index(Source):
    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskIndex(*args, **kwargs)

    @abstractmethod
    def __len__(self):
        self._not_implemented()

    def _normalize_kwargs_names(self, **kwargs):
        return kwargs

    def sel(self, *args, remapping=None, **kwargs):
        """Filter elements on their metadata(), according to kwargs.
        Returns a new index object.
        """

        kwargs = normalize_selection(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)
        if not kwargs:
            return self

        selection = Selection(kwargs, remapping=remapping)

        indices = (
            i for i, element in enumerate(self) if selection.match_element(element)
        )

        return self.new_mask_index(self, indices)

    def order_by(self, *args, remapping=None, patches=None, **kwargs):
        """Default order_by method.
        It expects that calling self[i] returns an element that and Order object can rank
        (i.e. order.get_element_ranking(element) -> tuple).
        then it sorts the elements according to the tuples.

        Returns a new index object.
        """
        kwargs = normalize_order_by(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)

        remapping = build_remapping(remapping, patches)

        if not kwargs:
            return self

        order = Order(kwargs, remapping=remapping)

        def cmp(i, j):
            return order.compare_elements(self[i], self[j])

        indices = list(range(len(self)))
        indices = sorted(indices, key=functools.cmp_to_key(cmp))
        return self.new_mask_index(self, indices)

    def __getitem__(self, n):
        if isinstance(n, slice):
            return self.from_slice(n)
        if isinstance(n, tuple):
            return self.from_tuple(n)
        if isinstance(n, list):
            return self.from_mask(n)
        if isinstance(n, dict):
            return self.from_dict(n)
        return self._getitem(n)

    def from_slice(self, s):
        indices = range(len(self))[s]
        return self.new_mask_index(self, indices)

    def from_mask(self, lst):
        indices = [i for i, x in enumerate(lst) if x]
        return self.new_mask_index(self, indices)

    def from_tuple(self, lst):
        return self.new_mask_index(self, lst)

    def from_dict(self, dic):
        return self.sel(dic)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, Index) for _ in sources)
        return MultiIndex(sources)

    def to_numpy(self, *args, **kwargs):
        import numpy as np

        return np.array([f.to_numpy(*args, **kwargs) for f in self])

    def to_pytorch_tensor(self, *args, **kwargs):
        import torch

        return torch.Tensor(self.to_numpy(*args, **kwargs))

    def full(self, *coords):
        return FullIndex(self, *coords)


class MaskIndex(Index):
    def __init__(self, index, indices):
        self.index = index
        self.indices = list(indices)

    def _getitem(self, n):
        n = self.indices[n]
        return self.index[n]

    def __iter__(self):
        return (self.index[i] for i in self.indices)

    def __len__(self):
        return len(self.indices)

    def __repr__(self):
        return "MaskIndex(%r,len=%s)" % (self.index, len(self.indices))


class MultiIndex(Index):
    def __init__(self, indexes, *args, **kwargs):
        self.indexes = list(indexes)
        super().__init__(*args, **kwargs)
        # self.indexes = list(i for i in indexes if len(i))
        # TODO: propagate  index._init_args, index._init_order_by, index._init_kwargs, for each i in indexes?

    def sel(self, *args, **kwargs):
        if not args and not kwargs:
            return self
        return self.__class__(i.sel(*args, **kwargs) for i in self.indexes)

    def _getitem(self, n):
        k = 0
        while n >= len(self.indexes[k]):
            n -= len(self.indexes[k])
            k += 1
        return self.indexes[k][n]

    def __len__(self):
        return sum(len(i) for i in self.indexes)

    def graph(self, depth=0):
        print(" " * depth, self.__class__.__name__)
        for s in self.indexes:
            s.graph(depth + 3)

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ",".join(repr(i) for i in self.indexes),
        )


class FullIndex(Index):
    def __init__(self, index, *coords):
        import numpy as np

        self.index = index

        # Pass1, unique values
        unique = index.unique_values(*coords)
        shape = tuple(len(v) for v in unique.values())

        name_to_index = defaultdict(dict)

        for k, v in unique.items():
            for i, e in enumerate(v):
                name_to_index[k][e] = i

        self.size = math.prod(shape)
        self.shape = shape
        self.holes = np.full(shape, False)

        for f in index:
            idx = tuple(name_to_index[k][f.metadata(k)] for k in coords)
            self.holes[idx] = True

        self.holes = self.holes.flatten()
        print("+++++++++", self.holes.shape, coords, self.shape)

    def __len__(self):
        return self.size

    def _getitem(self, n):
        assert self.holes[n], f"Attempting to access hole {n}"
        return self.index[sum(self.holes[:n])]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import warnings
from itertools import zip_longest
from numbers import Number

import numpy as np

from climetlab.core import Base

LOG = logging.getLogger(__name__)


# def to_tfdataset(
#     features,
#     targets=None,
#     total_size=None,
#     num_parallel_calls=10,
#     prefetch=1024,
#     **kwargs,
# ):
#
#     import tensorflow as tf
#
#     if features is not None and not callable(features):
#         if total_size is None:
#             LOG.debug("No total_size specified, infering from features.")
#             total_size = len(features)
#
#         _features = features
#
#         def features(i):
#             return _features[i].to_numpy()
#
#     if targets is not None and not callable(targets):
#         _targets = targets
#
#         def targets(i):
#             return _targets[i].to_numpy()
#
#     assert total_size is not None
#
#     def map_fn(i):
#         i = int(i)
#         return features(i)
#
#     @tf.function
#     def tf_map_fn(i):
#         return tf.py_function(func=map_fn, inp=[i], Tout=tf.float32)
#
#     def map_label_fn(i):
#         i = int(i)
#         return targets(i)
#
#     @tf.function
#     def tf_map_label_fn(i):
#         return tf.py_function(func=map_label_fn, inp=[i], Tout=tf.float32)
#
#     def dataset(mapping):
#         return (
#             tf.data.Dataset.range(total_size)
#             .map(mapping, num_parallel_calls=num_parallel_calls)
#             .prefetch(prefetch)
#         )
#
#     if targets is None:
#         return dataset(tf_map_fn)
#
#     return tf.data.Dataset.zip(
#         (
#             dataset(tf_map_fn),
#             dataset(tf_map_label_fn),
#         )
#     )


def default_merger(*funcs):
    # old: moved to climetlab.ml
    if not funcs:
        return None

    def map_fn(i):
        i = int(i)
        arrays = [m(i) for m in funcs]
        array = np.stack(arrays)
        return array

    return map_fn


def as_numpy_func(ds, options=None):
    # old: moved to climetlab.ml
    if ds is None or callable(ds):
        return ds

    def _options(new):
        o = {k: v for k, v in ds.get_options().items()}
        if new:
            o.update(new)
        return o

    options = _options(options)

    to_numpy_kwargs = options.get("to_numpy_kwargs", {})

    def take_i(i):
        return ds[i].to_numpy(**to_numpy_kwargs)

    if "offset" in options and options["offset"]:
        offset = options["offset"]

        def take_i(i):  # noqa: F811
            return ds[i + offset].to_numpy(**to_numpy_kwargs)

    func = take_i

    if "constant" in options and options["constant"]:

        def first(func):
            def wrap(i):
                return func(0)

            return wrap

        func = first(func)

    if "normalize" in options:
        a, b = normalize_a_b(options["normalize"], ds)

        def normalize(func):
            def wrap(i):
                return a * func(i) + b

            return wrap

        func = normalize(func)

    return func


def normalize_a_b(option, dataset):
    if isinstance(option, (tuple, list)) and all(
        [isinstance(x, Number) for x in option]
    ):
        a, b = option
        return a, b

    if option == "mean-std":
        stats = dataset.statistics()
        average, stdev = stats["average"], stats["stdev"]
        if stdev < (average * 1e-6):
            warnings.warn(
                f"Normalizing: the field seems to have only one value {stats}"
            )
        return 1 / stdev, -average / stdev

    if option == "min-max":
        stats = dataset.statistics()
        mini, maxi = stats["minimum"], stats["maximum"]
        x = maxi - mini
        if x < 1e-9:
            warnings.warn(
                f"Normalizing: the field seems to have only one value {stats}."
            )
        return 1 / x, -mini / x

    raise ValueError(option)


def to_funcs(features, targets, options, targets_options, merger, targets_merger):
    if targets is None:
        targets = []

    if options is None:
        options = [{} for i in features]

    if targets_options is None:
        targets_options = [{} for i in targets]

    assert isinstance(features, (list, tuple)), features
    assert len(features) == len(options), (len(features), len(options))
    funcs = [
        as_numpy_func(_, opt) for _, opt in zip_longest(features, options, fillvalue={})
    ]
    funcs_targets = [
        as_numpy_func(_, opt)
        for _, opt in zip_longest(targets, targets_options, fillvalue={})
    ]

    func = merger(*funcs)
    func_targets = targets_merger(*funcs_targets)

    return func, func_targets


def to_tfdataset2(
    total_size=None,
    features=None,
    targets=None,
    options=None,
    targets_options=None,
    merger=default_merger,
    targets_merger=default_merger,
    #
    num_parallel_calls=10,
    prefetch=1024,
    shuffle_buffer_size=100,
    **kwargs,
):
    if total_size is None:
        total_size = len(features[0])

    import tensorflow as tf

    func, func_targets = to_funcs(
        features, targets, options, targets_options, merger, targets_merger
    )

    indices = tf.data.Dataset.range(total_size)
    if shuffle_buffer_size:
        indices = indices.shuffle(shuffle_buffer_size)

    def dataset(f):
        def map_fn(i):
            i = int(i)
            return f(i)

        @tf.function  # only for debugging
        def tf_map_fn(i):
            return tf.py_function(func=map_fn, inp=[i], Tout=tf.float32)

        ds = indices.map(tf_map_fn, num_parallel_calls=num_parallel_calls)
        ds = ds.prefetch(prefetch)
        return ds

    if func_targets is None:
        tfds = dataset(func)
        tfds._climetlab_shape = func(0).shape
        return tfds

    tf_features = dataset(func)
    tf_targets = dataset(func_targets)
    tfds = tf.data.Dataset.zip((tf_features, tf_targets))

    # TODO: use metadata from grib directly, without loading the numpy array
    tfds._climetlab_tf_shape_in = func(0).shape

    if targets:
        tfds._climetlab_tf_shape_out = func_targets(0).shape

    tfds._climetlab_tf_input = tf_features
    if targets:
        tfds._climetlab_tf_targets = tf_targets
    # ds = ds.prefetch(prefetch)

    return tfds


class NumpyFuncWrapper:
    __slots__ = ["_wrapped_ds", "_wrapped_opt"]

    def __init__(self, ds, opt={}):
        self.__class__._wrapped_ds.__set__(self, ds)
        self.__class__._wrapped_opt.__set__(self, opt)

    def __getattr__(self, name):
        return getattr(self._wrapped_ds, name)

    def __setattr__(self, name, value):
        return setattr(self._wrapped_ds, name, value)


class TensorflowMixIn:
    # def to_tfdataset(
    #     self,
    #     labels=None,
    #     targets=None,
    #     **kwargs,
    # ):
    #     if targets is None:  # rename "labels" into "targets" ?
    #         targets = labels
    #     return to_tfdataset(features=self, targets=targets, **kwargs)

    def to_tfdataset(self, *args, **kwargs):
        if "features" not in kwargs:
            kwargs["features"] = [self]

        if "total_size" not in kwargs:
            kwargs["total_size"] = len(self)

        return to_tfdataset2(*args, **kwargs)

    def to_tfdataset_(
        self, *others, align_with=None, total_size=None, merger=default_merger
    ):
        import tensorflow as tf

        if align_with is not None:
            indices = align_with._climetlab_indices
        else:
            if total_size is None:
                total_size = len(self)
            indices = tf.data.Dataset.range(total_size)

        # shuffle can happen early
        # indices = indices.shuffle(shuffle)

        def as_numpy_func(ds):
            if isinstance(ds, Base):
                func = ds._to_numpy_func()
                ds._numpy_func = func
                return func
            if hasattr(ds, "_climetlab_numpy_func"):
                return ds._climetlab_numpy_func
            raise ValueError(str(ds))

        funcs = [as_numpy_func(ds) for ds in [self, *others]]
        func = merger(*funcs)

        tfds = indices.map(to_tf_func(func))

        tfds._climetlab_n = len(others) + 1
        tfds._climetlab_tf_shape = (tfds._climetlab_n, *self[0].shape)
        tfds._climetlab_indices = indices
        tfds._climetlab_numpy_func = func
        return tfds

    def _to_tf_func(self, opt={}):
        import tensorflow as tf

        func = self._to_numpy_func(opt)

        @tf.function  # only for debugging
        def tf_map_fn(i):
            return tf.py_function(func=func, inp=[i], Tout=tf.float32)

        return tf_map_fn

    def _to_tfdataset(self, indices, opt={}):
        tfds = indices.map(self._to_tf_func(opt))
        tfds._climetlab_numpy_func = self._numpy_func
        return tfds

    def zip_with(self, *args, **kwargs):
        import tensorflow as tf

        return tf.data.Dataset.zip(tuple(self, *args), **kwargs)


def cml_tfzip(*args, **kwargs):
    import tensorflow as tf

    return tf.data.Dataset.zip(tuple(args), **kwargs)


def to_tf_func(map_fn):
    import tensorflow as tf

    @tf.function  # only for debugging
    def tf_map_fn(i):
        return tf.py_function(func=map_fn, inp=[i], Tout=tf.float32)

    return tf_map_fn



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from functools import cached_property
from itertools import product

from climetlab.core.index import MaskIndex, MultiIndex
from climetlab.indexing.fieldset import FieldSet
from climetlab.utils.bbox import BoundingBox
from climetlab.utils.dates import to_datetime

from .coords import LevelCoordinate, OtherCoordinate, TimeCoordinate
from .dataset import DataSet
from .field import NetCDFField


class NetCDFFieldSet(FieldSet):
    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return NetCDFMaskFieldSet(*args, **kwargs)

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __getitem__(self, n):
        return self.fields[n]

    @cached_property
    def fields(self):
        return self._get_fields(DataSet(self.xr_dataset))

    def _get_fields(self, ds):  # noqa C901
        # Select only geographical variables
        has_lat = False
        has_lon = False

        fields = []

        skip = set()

        for name in ds.data_vars:
            v = ds[name]
            skip.update(getattr(v, "coordinates", "").split(" "))
            skip.update(getattr(v, "bounds", "").split(" "))
            skip.update(getattr(v, "grid_mapping", "").split(" "))

        for name in ds.data_vars:
            if name in skip:
                continue

            v = ds[name]

            coordinates = []

            # self.log.info('Scanning file: %s var=%s coords=%s', self.path, name, v.coords)

            info = [value for value in v.coords if value not in v.dims]
            non_dim_coords = {}
            for coord in v.coords:
                if coord not in v.dims:
                    non_dim_coords[coord] = ds[coord].values
                    continue

                c = ds[coord]

                # self.log.info("COORD %s %s %s %s", coord, type(coord), hasattr(c, 'calendar'), c)

                standard_name = getattr(c, "standard_name", None)
                axis = getattr(c, "axis", None)
                long_name = getattr(c, "long_name", None)

                use = False

                if (
                    standard_name in ("longitude", "projection_x_coordinate")
                    or (long_name == "longitude")
                    or (axis == "X")
                ):
                    has_lon = True
                    use = True

                if (
                    standard_name in ("latitude", "projection_y_coordinate")
                    or (long_name == "latitude")
                    or (axis == "Y")
                ):
                    has_lat = True
                    use = True

                # Of course, not every one sets the standard_name
                if standard_name in ("time", "forecast_reference_time") or axis == "T":
                    coordinates.append(TimeCoordinate(c, coord in info))
                    use = True

                # TODO: Support other level types
                if standard_name in (
                    "air_pressure",
                    "model_level_number",
                    "altitude",
                ):  # or axis == 'Z':
                    coordinates.append(LevelCoordinate(c, coord in info))
                    use = True

                if axis in ("X", "Y"):
                    use = True

                if not use:
                    coordinates.append(OtherCoordinate(c, coord in info))

            if not (has_lat and has_lon):
                # self.log.info("NetCDFReader: skip %s (Not a 2 field)", name)
                continue

            for values in product(*[c.values for c in coordinates]):
                slices = []
                for value, coordinate in zip(values, coordinates):
                    slices.append(coordinate.make_slice(value))

                fields.append(NetCDFField(self, ds, name, slices, non_dim_coords))

        if not fields:
            raise Exception("NetCDFReader no 2D fields found in %s" % (self.path,))

        return fields

    def to_xarray(self, **kwargs):
        import xarray as xr

        if self.path.startswith("http"):
            return xr.open_dataset(self.path, **kwargs)
        return type(self).to_xarray_multi_from_paths([self.path], **kwargs)

    @classmethod
    def to_xarray_multi_from_paths(cls, paths, **kwargs):
        import xarray as xr

        options = dict()
        options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))

        return xr.open_mfdataset(
            paths,
            **options,
        )

    def plot_map(self, *args, **kwargs):
        return self.fields[0].plot_map(*args, **kwargs)

    # Used by normalisers
    def to_datetime(self):
        times = self.to_datetime_list()
        assert len(times) == 1
        return times[0]

    def to_datetime_list(self):
        # TODO: check if that can be done faster
        result = set()
        for s in self.fields:
            result.add(to_datetime(s.time))
        return sorted(result)

    def to_bounding_box(self):
        return BoundingBox.multi_merge([s.to_bounding_box() for s in self.fields])

    @classmethod
    def merge(cls, sources):
        assert len(sources) > 1
        assert all(isinstance(_, NetCDFFieldSet) for _ in sources)
        return NetCDFMultiFieldSet(sources)


class NetCDFFieldSetFromFileOrURL(NetCDFFieldSet):
    def __init__(self, path_or_url):
        self.path_or_url = path_or_url

    @cached_property
    def xr_dataset(self):
        import xarray as xr

        return xr.open_dataset(self.path_or_url)


class NetCDFFieldSetFromFile(NetCDFFieldSetFromFileOrURL):
    def __init__(self, path):
        super().__init__(path)

    def __repr__(self):
        return "NetCDFFieldSetFromFile(%s)" % (self.path_or_url,)

    def to_metview(self):
        from climetlab.metview import mv_read

        return mv_read(self.path_or_url)


class NetCDFFieldSetFromURL(NetCDFFieldSetFromFileOrURL):
    def __init__(self, url):
        super().__init__(url)

    def __repr__(self):
        return "NetCDFFieldSetFromURL(%s)" % (self.path_or_url,)


class NetCDFMaskFieldSet(NetCDFFieldSet, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)
        self.path = "<mask>"

    @cached_property
    def fields(self):
        return list(self.index[i] for i in self.indices)


class NetCDFMultiFieldSet(NetCDFFieldSet, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)
        self.paths = [s.path for s in args[0]]
        self.path = "<multi>"

    def to_xarray(self, **kwargs):
        import xarray as xr

        if not kwargs:
            kwargs = dict(combine="by_coords")
        return xr.open_mfdataset(self.paths, **kwargs)

    @cached_property
    def fields(self):
        result = []
        for s in self.indexes:
            result.extend(s.fields)
        return result

    def __len__(self):
        return MultiIndex.__len__(self)

    def __getitem__(self, n):
        return MultiIndex.__getitem__(self, n)



#!/usr/bin/env python
#
# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#
import itertools
import logging

from climetlab.utils.humanize import did_you_mean

LOG = logging.getLogger(__name__)


SYNONYMS = (
    (("mars", "2t"), ("cf", "t2m")),
    (("mars", "ci"), ("cf", "siconc")),
)

VOCABULARIES = {}

SYNONYMS = (
    (("mars", "2t"), ("cf", "t2m")),
    (("mars", "ci"), ("cf", "siconc")),
    (("mars", "sst"), ("cf", "sst")),
    (("mars", "rsn"), ("cf", "rsn")),
    (("mars", "tcc"), ("cf", "tcc")),
    (("mars", "tcw"), ("cf", "tcw")),
    (("mars", "sm20"), ("cf", "sm20")),
    (("mars", "sm100"), ("cf", "sm100")),
    (("mars", "st20"), ("cf", "st20")),
    (("mars", "st100"), ("cf", "st100")),
    (("mars", "tp"), ("cf", "tp")),
    (("mars", "ttr"), ("cf", "ttr")),
    (("mars", "sp"), ("cf", "sp")),
    (("mars", "msl"), ("cf", "msl")),
    (("mars", "lsm"), ("cf", "lsm")),
    (("mars", "u"), ("cf", "u")),
    (("mars", "v"), ("cf", "v")),
    (("mars", "gh"), ("cf", "gh")),
    (("mars", "t"), ("cf", "t")),
    (("mars", "q"), ("cf", "q")),
)


class Vocabulary:
    def __init__(self, name):
        self.name = name
        self.words = set()
        self.aliases = {}

    def add(self, word, *aliases):
        self.words.add(word)
        for a in aliases:
            self.aliases[a] = word

    def lookup(self, word):
        while word in self.aliases:
            word = self.aliases[word]

        return word if word in self.words else None

    def normalise(self, word):
        w = self.lookup(word)
        if w is not None:
            return w

        # For now....
        for synonyms in SYNONYMS:
            matches = [s for s in synonyms if s[0] != self.name and s[1] == word]
            if not matches:
                continue
            assert len(matches) == 1, matches
            for name, w in synonyms:
                if name == self.name:
                    return w

        correction = did_you_mean(
            word,
            itertools.chain(
                self.words,
                self.aliases.keys(),
            ),
        )
        if correction is not None:
            LOG.debug(
                "This warning can be safely ignored: Cannot find '%s' in %s vocabulary, did you mean '%s'?",
                word,
                self.name,
                correction,
            )

        return word


mars = Vocabulary("mars")
mars.add("2t")
mars.add("tp")
mars.add("ci")
VOCABULARIES["mars"] = mars

cf = Vocabulary("cf")
cf.add("t2m")
cf.add("tp")
cf.add("siconc")
VOCABULARIES["cf"] = cf



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import functools
import inspect
import logging
import os
import re
import threading

from climetlab.utils import load_json_or_yaml
from climetlab.utils.availability import Availability

LOG = logging.getLogger(__name__)


def dict_args(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        m = []
        p = {}
        for q in args:
            if isinstance(q, dict):
                p.update(q)
            else:
                m.append(q)
        p.update(kwargs)
        return func(*m, **p)

    return wrapped


LOCK = threading.RLock()


def locked(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        with LOCK:
            return func(*args, **kwargs)

    return wrapped


class Decorator:
    is_availability = False

    def __call__(self, func):
        from climetlab.arguments import InputManager

        if not callable(func):
            manager = InputManager(decorators=[self])
            return manager.apply_to_value(func)

        decorators = [self]

        def unwrap(f):
            if not hasattr(f, "_climetlab_decorators"):
                return f
            return unwrap(f.__wrapped__)

        unwrapped = unwrap(func)

        if hasattr(func, "_climetlab_decorators"):
            decorators = decorators + func._climetlab_decorators

        manager = InputManager(decorators=decorators)

        @functools.wraps(unwrapped)
        def newfunc(*args, **kwargs):
            args, kwargs = manager.apply_to_arg_kwargs(args, kwargs, func=unwrapped)
            return unwrapped(*args, **kwargs)

        newfunc._climetlab_decorators = decorators

        return newfunc


OPTIONS = {
    "date": ("format",),
    "date-list": ("format",),
    "bounding-box": ("format",),
    "bbox": ("format",),
    "variable": ("convention",),
    "variable-list": ("convention",),
}


def normalize_grib_keys(f):
    f = alias_argument("levelist", ["level", "levellist"])(f)
    f = alias_argument("levtype", ["leveltype"])(f)
    f = alias_argument("param", ["variable", "parameter"])(f)
    f = alias_argument("number", ["realization", "realisation"])(f)
    f = alias_argument("class", ["klass", "class_"])(f)
    return f


def _normalize_time_as_tuple(time, type):
    if isinstance(time, str):
        time = (time,)
    return tuple(_normalize_time(t, type) for t in time)


def _normalize_time(time, type):
    assert type in (int, str), type

    if time is None or time == all:
        return time

    if isinstance(time, list):
        return [_normalize_time(t, type) for t in time]
    if isinstance(time, tuple):
        return tuple([_normalize_time(t, type) for t in time])

    try:
        time = int(time)
    except ValueError:
        return time

    if time % 100:  # not multiple of 100
        time = time * 100

    assert time <= 2400, time
    assert time >= 0, time

    if type is str:
        return f"{time:04d}"
    assert isinstance(time, int)
    return time


def _normalize_expver(expver):
    if isinstance(expver, str):
        return expver
    assert isinstance(expver, int)
    return f"{expver:04}"


def _normalize_expver_as_tuple(expver):
    if isinstance(expver, (int, str)):
        expver = _normalize_expver(expver)
        expver = (expver,)
    lst = []
    for x in expver:
        if not isinstance(x, str):
            x = "%04d" % x
        lst.append(x)
    return tuple(lst)


def normalize_grib_key_values(kwargs, accept_none=True, as_tuple=False):
    def f(**kwargs):
        return kwargs

    f = normalize_grib_keys(f)
    f = normalize("param", "variable-list(mars)")(f)
    f = normalize("date", "date-list(%Y%m%d)")(f)
    f = normalize("hdate", "date-list(%Y%m%d)")(f)
    f = normalize("area", "bounding-box(list)")(f)
    f = normalize("levelist", "int-list")(f)
    kwargs = f(**kwargs)

    if "time" in kwargs:
        kwargs["time"] = {False: _normalize_time, True: _normalize_time_as_tuple}[
            as_tuple
        ](kwargs["time"], int)
    if "expver" in kwargs:
        kwargs["expver"] = {False: _normalize_expver, True: _normalize_expver_as_tuple}[
            as_tuple
        ](kwargs["expver"])

    return kwargs


class alias_argument(Decorator):
    def __init__(
        self,
        target=None,
        aliases=None,
        **kwargs,
    ):
        """target: actual argument name in the function to decorate.
        aliases: str or list of str to create aliases to target.
        """
        if target is not None:
            assert aliases is not None
            assert not kwargs
            kwargs = {target: aliases}

        for k, v in kwargs.items():
            if isinstance(v, (list, tuple)):
                continue
            if isinstance(v, str):
                kwargs[k] = [v]
                continue
            if isinstance(v, dict):
                raise ValueError(
                    (
                        "Error: alias_argument is expecting a list or str."
                        f" You may be looking for: @normalize(aliases={kwargs})"
                    )
                )
            raise ValueError(f"Wrong alias list for '{k}':{v}")

        self.kwargs = kwargs

        # check for duplicate aliases
        for k, v in self.kwargs.items():
            for k_, v_ in self.kwargs.items():
                if k == k_:
                    continue
                intersection = set(v).intersection(set(v_))
                if intersection:
                    raise ValueError(
                        f"Error: alias_argument cannot alias '{list(intersection)[0]}' to '{k}' and '{k_}'"
                    )

    def register(self, manager):
        manager.register_alias_argument(self)


class normalize(Decorator):
    def __init__(
        self,
        name,
        values=None,
        **kwargs,
    ):
        assert name is None or isinstance(name, str)
        self.name = name

        if isinstance(values, str):
            assert (
                kwargs.get("type") is None
            ), f"Cannot mix values={values} and type={kwargs.get('type')}"
            if "(" in values:
                m = re.match(r"(.+)\((.+)\)", values)
                type = m.group(1)
                args = m.group(2).split(",")
            else:
                type = values
                args = []

            # len(args) <= len(options)
            if args:
                for name, value in zip(OPTIONS[type], args):
                    kwargs[name] = value
            kwargs["type"] = type
        else:
            kwargs["values"] = values

        if "aliases" in kwargs and isinstance(kwargs["aliases"], str):
            _, ext = os.path.splitext(kwargs["aliases"])
            if ext in (".json", ".yaml", ".yml"):
                path = kwargs["aliases"]
                if not os.path.isabs(path):
                    caller = os.path.dirname(inspect.stack()[1].filename)
                    path = os.path.join(caller, path)
                kwargs["aliases"] = load_json_or_yaml(path)

        self.kwargs = kwargs

    def register(self, manager):
        manager.register_normalize(self)


class availability(Decorator):
    is_availability = True

    def __init__(self, availability, **kwargs):
        if isinstance(availability, str):
            if not os.path.isabs(availability):
                caller = os.path.dirname(inspect.stack()[1].filename)
                availability = os.path.join(caller, availability)

        self.availability = Availability(availability, **kwargs)

    def register(self, manager):
        manager.register_availability(self)


def cached_method(method):
    name = f"_{method.__name__}"

    @functools.wraps(method)
    def wrapped(self):
        if getattr(self, name, None) is None:
            setattr(self, name, method(self))
        return getattr(self, name)

    return wrapped



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from climetlab.utils import module_loaded
from climetlab.wrappers import get_wrapper


def wrapper(data, *args, **kwargs):
    if not module_loaded("torch"):
        return None

    import torch  # noqa

    if isinstance(data, torch.Tensor):
        return get_wrapper(data.detach().cpu().numpy(), *args, **kwargs)

    return None
